"""Quasi-bi-clique and the densest sub-binary matrix problem main module."""

# Due to typer usage:
# ruff: noqa: TCH001, TCH003, UP007, FBT001, PLR0913

from __future__ import annotations

import logging
import sys
from pathlib import Path
from typing import Optional

import typer

import qbc_dsbm.binary_matrix.apps.create as bm_apps_create
import qbc_dsbm.binary_matrix.apps.ops as bm_apps_ops
import qbc_dsbm.binary_matrix.apps.views as bm_apps_views
from qbc_dsbm import QBCExactModels, QBCHeuristicModels
from qbc_dsbm.binary_matrix.items import BinMatrix
from qbc_dsbm.cli import (
    CommonArgs,
    DivCqrArgs,
    SolverChoice,
    UniqueArgs,
    UniqueModelChoices,
)
from qbc_dsbm.io_config import IOConfig, IOConfigDivCqr
from qbc_dsbm.loggers import DivCqrStrategyLogger, UniqueStrategyLogger
from qbc_dsbm.models.solvers import CBCSolver, GurobiSolver, Solver
from qbc_dsbm.strategies import (
    DivideAndConquerStrategyError,
    UniqueStrategyError,
    divide_and_conquer_strategy,
    unique_strategy,
)

__APP = typer.Typer()


_LOGGER = logging.getLogger("qbc_dsbm")
_LOGGER.setLevel(logging.DEBUG)


__APP.add_typer(bm_apps_create.APP, name="bm_create", help="Binary matrix creation")
__APP.add_typer(bm_apps_ops.APP, name="bm_ops", help="Binary matrix operations")
__APP.add_typer(bm_apps_views.APP, name="bm_views", help="Binary matrix views")


@__APP.command()
def unique(
    csv_path: Path = CommonArgs.ARG_CSV_PATH,
    solver: SolverChoice = CommonArgs.OPT_SOLVER,
    solve_time_lim: Optional[int] = CommonArgs.OPT_SOLVE_TIME_LIM,
    solver_log: bool = CommonArgs.OPT_SOLVER_PRINT_LOG,
    minimum_number_of_rows: int = UniqueArgs.OPT_MIN_NUMBER_OF_ROWS,
    minimum_number_of_columns: int = UniqueArgs.OPT_MIN_NUMBER_OF_COLUMNS,
    check_satisfiability: bool = UniqueArgs.OPT_CHECK_SATISFIABILITY,
    output_directory: Path = CommonArgs.OPT_OUTDIR,
    model: UniqueModelChoices = UniqueArgs.ARG_MODEL,
    epsilon: float = UniqueArgs.OPT_EPSILON,
    debug: bool = CommonArgs.OPT_DEBUG,
) -> None:
    """Solve with one model."""
    __format_logger(debug)
    __init_output_directory(output_directory)
    solver_obj = __create_solver_obj(solver, solve_time_lim, solver_log)
    io_config = IOConfig(
        csv_path,
        model.to_unique_model(),
        epsilon,
        solver_obj,
        minimum_number_of_rows,
        minimum_number_of_columns,
        check_satisfiability,
    )
    try:
        result_bin_matrix, logger = unique_strategy(io_config)
    except UniqueStrategyError as error:
        error.logger().write_yaml(output_directory)
        _LOGGER.critical(str(error))
        sys.exit(1)
    else:
        __finalize(result_bin_matrix, logger, output_directory)


@__APP.command()
def divcqr(
    csv_path: Path = CommonArgs.ARG_CSV_PATH,
    solver: SolverChoice = CommonArgs.OPT_SOLVER,
    solve_time_lim: Optional[int] = CommonArgs.OPT_SOLVE_TIME_LIM,
    solver_log: bool = CommonArgs.OPT_SOLVER_PRINT_LOG,
    output_directory: Path = CommonArgs.OPT_OUTDIR,
    heuristic_model: QBCHeuristicModels = DivCqrArgs.ARG_HEURISTIC_MODEL,
    heuristic_epsilon: float = DivCqrArgs.ARG_HEURISTIC_EPSILON,
    exact_model: QBCExactModels = DivCqrArgs.ARG_EXACT_MODEL,
    exact_epsilon: float = DivCqrArgs.ARG_EXACT_EPSILON,
    debug: bool = CommonArgs.OPT_DEBUG,
) -> None:
    """Solve with the divide and conquer strategy."""
    __format_logger(debug)
    __init_output_directory(output_directory)
    solver_obj = __create_solver_obj(solver, solve_time_lim, solver_log)
    io_config = IOConfigDivCqr(
        csv_path,
        heuristic_model,
        heuristic_epsilon,
        exact_model,
        exact_epsilon,
        solver_obj,
    )
    try:
        result_bin_matrix, logger = divide_and_conquer_strategy(io_config)
    except DivideAndConquerStrategyError as error:
        error.logger().write_yaml(output_directory)
        _LOGGER.critical(str(error))
    else:
        __finalize(result_bin_matrix, logger, output_directory)


def __format_logger(debug: bool) -> None:
    """Format logger."""
    handler = logging.StreamHandler(sys.stdout)
    if debug:
        handler.setLevel(logging.DEBUG)
    else:
        handler.setLevel(logging.INFO)
    formatter = logging.Formatter(
        "%(asctime)s %(name)s [%(levelname)s] %(message)s",
    )
    handler.setFormatter(formatter)
    _LOGGER.addHandler(handler)
    log_filter = logging.Filter("tilling")
    _LOGGER.addFilter(log_filter)


def __init_output_directory(output_directory: Path) -> None:
    try:
        output_directory.mkdir(parents=True, exist_ok=False)
    except FileExistsError:
        _LOGGER.critical("Output directory already exists: %s", output_directory)
        sys.exit(1)


def __create_solver_obj(
    solver: SolverChoice,
    solve_time_lim: int | None,
    solver_log: bool,
) -> Solver:
    match solver:
        case SolverChoice.CBC:
            return CBCSolver(
                time_limit=solve_time_lim,
                print_log=solver_log,
            )
        case SolverChoice.GUROBI:
            return GurobiSolver(
                time_limit=solve_time_lim,
                print_log=solver_log,
            )
        case _:
            _LOGGER.critical("Unknown solver: %s", solver)
            sys.exit(1)


def __finalize(
    result_bin_matrix: BinMatrix,
    logger: DivCqrStrategyLogger | UniqueStrategyLogger,
    output_directory: Path,
) -> None:
    logger.write_yaml(output_directory)
    result_bin_matrix.to_csv(output_directory / "result_bin_matrix.csv")


def main() -> None:
    """Execute the typer app."""
    __APP()


if __name__ == "__main__":
    main()



"""CLI module."""

from __future__ import annotations

from contextlib import suppress
from dataclasses import dataclass
from enum import Enum
from pathlib import Path

import typer

from qbc_dsbm import BCModels, QBCExactModels, QBCHeuristicModels, UniqueModels


class SolverChoice(Enum):
    """Solver choice."""

    GUROBI = "Gurobi"
    CBC = "CBC"


class UniqueModelChoices(Enum):
    """Unique model choices."""

    MAX_ONES = QBCExactModels.MAX_ONES.value
    MAX_SURFACE = QBCExactModels.MAX_SURFACE.value
    MAX_PERIMETER = QBCExactModels.MAX_PERIMETER.value
    MAX_ONES_COMPACT = QBCExactModels.MAX_ONES_COMPACT.value

    KONIG_V = BCModels.KONIG_V.value
    KONIG_E = BCModels.KONIG_E.value

    MIN_DEL_RC = QBCHeuristicModels.MIN_DEL_RC.value
    MIN_DEL_ROWS = QBCHeuristicModels.MIN_DEL_ROWS.value
    MIN_DEL_COLS = QBCHeuristicModels.MIN_DEL_COLS.value
    MIN_DEL_ROWS_RELAX = QBCHeuristicModels.MIN_DEL_ROWS_RELAX.value
    MIN_DEL_COLS_RELAX = QBCHeuristicModels.MIN_DEL_COLS_RELAX.value
    MIN_DEL_ONES = QBCHeuristicModels.MIN_DEL_ONES.value
    KP_QB = QBCHeuristicModels.KP_QB.value

    def to_unique_model(self) -> UniqueModels:
        """Convert to unique model.

        Returns
        -------
        UniqueModels
            Unique model

        """
        with suppress(ValueError):
            return QBCExactModels(self.value)
        with suppress(ValueError):
            return BCModels(self.value)
        return QBCHeuristicModels(self.value)


# ============================================================================ #
#                                   CONSTANTS                                  #
# ============================================================================ #
#
# Common args/opts
#
@dataclass
class CommonArgs:
    """Common args and opts."""

    ARG_CSV_PATH = typer.Argument(parser=Path, help="Binary matrix CSV file path")

    OPT_SOLVER_DEF = SolverChoice.CBC
    OPT_SOLVER = typer.Option(
        default=OPT_SOLVER_DEF.value,
        help="The solver to use",
    )

    OPT_SOLVE_TIME_LIM_DEF: int | None = None
    OPT_SOLVE_TIME_LIM = typer.Option(
        default=OPT_SOLVE_TIME_LIM_DEF,
        help="Solve time limit in seconds",
    )

    OPT_SOLVER_PRINT_LOG_DEF = False
    OPT_SOLVER_PRINT_LOG = typer.Option(
        default=OPT_SOLVER_PRINT_LOG_DEF,
        help="Print solver log",
    )

    OPT_OUTDIR_DEF = Path("./qbc_dsbm_results")
    OPT_OUTDIR = typer.Option(
        default=OPT_OUTDIR_DEF,
        help="The output directory path",
    )

    OPT_DEBUG_DEF = False
    OPT_DEBUG = typer.Option(
        default=OPT_DEBUG_DEF,
        help="Debug mode",
    )


#
# Command unique
#
@dataclass
class UniqueArgs:
    """Unique args and opts."""

    ARG_MODEL = typer.Argument(
        help="The model to use",
    )

    OPT_EPSILON_DEF = 0.0
    OPT_EPSILON = typer.Option(
        default=OPT_EPSILON_DEF,
        help="Remaining zeros percentage",
    )

    OPT_MIN_NUMBER_OF_ROWS = typer.Option(
        default=0,
        help="Minimum number of rows",
    )

    OPT_MIN_NUMBER_OF_COLUMNS = typer.Option(
        default=0,
        help="Minimum number of columns",
    )

    OPT_CHECK_SATISFIABILITY = typer.Option(
        default=False,
        help="Just check the satisfiability, do not solve an optimization problem",
    )


#
# Command divcqr
#
@dataclass
class DivCqrArgs:
    """DivCqr args and opts."""

    ARG_HEURISTIC_MODEL = typer.Argument(
        help="The heuristic model to use",
    )

    ARG_HEURISTIC_EPSILON = typer.Argument(
        help="The heuristic epsilon to use",
    )

    ARG_EXACT_MODEL = typer.Argument(
        help="The exact model to use",
    )

    ARG_EXACT_EPSILON = typer.Argument(
        help="The exact epsilon to use",
    )



"""Solvers module."""

from __future__ import annotations

import logging
import re
from abc import ABC, abstractmethod
from pathlib import Path
from typing import Any, cast

from pulp import (
    GUROBI,
    PULP_CBC_CMD,
    LpProblem,
    LpSolutionInfeasible,
    LpStatus,
    LpStatusNotSolved,
)

from qbc_dsbm import UniqueModels
from qbc_dsbm.models.views import (
    LPStats,
    ModelLPMILPStats,
    ModelLPStats,
    ModelSolveStats,
    WarmMILPStats,
)

_LOGGER = logging.getLogger(__name__)


# ============================================================================ #
#                                    CLASSES                                   #
# ============================================================================ #
class Solver(ABC):
    """Solver base class."""

    _NAME: str = "base solver class"
    KEY_NAME = "name"
    KEY_SOLVE_TIME_LIM = "time_limit"
    KEY_PRINT_LOG = "print_log"

    @classmethod
    def name(cls) -> str:
        """Return the solver name.

        Returns
        -------
        str
            Solver name

        """
        return cls._NAME

    @classmethod
    def from_dict(cls, solver_config_dict: dict[str, Any]) -> Solver:
        """Convert dict to Solver.

        Parameters
        ----------
        solver_config_dict : dict
            Dictionnary of solver config

        Returns
        -------
        Solver
            Solver

        """
        return cls(
            time_limit=solver_config_dict[cls.KEY_SOLVE_TIME_LIM],
            print_log=solver_config_dict[cls.KEY_PRINT_LOG],
        )

    def __init__(
        self,
        time_limit: float | None = None,
        *,
        print_log: bool = False,
    ) -> None:
        """Initialize solver."""
        self._time_limit = time_limit
        self._print_log = print_log

    def time_limit(self) -> float | None:
        """Return the time limit.

        Returns
        -------
        float | None
            Time limit

        """
        return self._time_limit

    def print_log(self) -> bool:
        """Return the print log flag.

        Returns
        -------
        bool
            Print log flag

        """
        return self._print_log

    @abstractmethod
    def solve_lp(
        self,
        prob: LpProblem,
        warm_start: bool = False,  # noqa: FBT001, FBT002
    ) -> ModelLPStats:
        """Solve LP with LP solver.

        Parameters
        ----------
        prob : LpProblem
            Problem to solve
        warm_start : bool, default=False
            Warm start

        Returns
        -------
        ModelLPStats
            LP statistics

        Raises
        ------
        NoSolutionError
            If no solution is found
        SolverTimeLimitReached
            If solver time limit is reached

        """
        raise NotImplementedError

    @abstractmethod
    def solve_milp(
        self,
        prob: LpProblem,
        warm_start: bool = False,  # noqa: FBT001, FBT002
    ) -> ModelLPMILPStats:
        """Solve problem with CBC.

        Parameters
        ----------
        prob : LpProblem
            Problem to solve
        warm_start : bool, default=False
            Warm start

        Returns
        -------
        MIPStats
            MIP statistics

        Raises
        ------
        NoSolutionError
            If no solution is found
        SolverTimeLimitReached
            If solver time limit is reached

        """
        raise NotImplementedError

    def to_dict(self) -> dict[str, Any]:
        """Convert to dict.

        Returns
        -------
        dict[str, Any]
            Dictionary

        """
        return {
            self.KEY_NAME: self._NAME,
            self.KEY_SOLVE_TIME_LIM: self._time_limit,
            self.KEY_PRINT_LOG: self._print_log,
        }


class CBCSolver(Solver):
    """CBC solver class."""

    _NAME = "CBC"
    __LOG_FILE_PATH = Path("tmp_cbc.log")

    def solve_lp(
        self,
        prob: LpProblem,
        warm_start: bool = False,  # noqa: FBT001, FBT002
    ) -> ModelLPStats:
        """Solve LP with CBC LP solver.

        Parameters
        ----------
        prob : LpProblem
            Problem to solve
        warm_start : bool, default=False
            Warm start

        Returns
        -------
        ModelLPStats
            LP statistics

        Raises
        ------
        NoSolutionError
            If no solution is found
        SolverTimeLimitReached
            If solver time limit is reached

        """
        if self._print_log:
            _LOGGER.info("Solve stats:")
        prob.solve(
            PULP_CBC_CMD(
                mip=False,
                timeLimit=self._time_limit,
                msg=self._print_log,
                warmStart=warm_start,
            ),
        )
        lp_stats = LPStats(
            LpStatus[prob.status],
            prob.solutionTime,
            prob.solutionCpuTime,
        )
        solve_stats = ModelLPStats(
            cast(UniqueModels, prob.name),
            prob.numVariables(),
            prob.numConstraints(),
            lp_stats,
        )
        if prob.status == LpSolutionInfeasible:
            raise NoSolutionError(solve_stats)
        if prob.status == LpStatusNotSolved:
            raise SolverTimeLimitReachedError(solve_stats)
        lp_stats.set_objective_value(
            float(prob.objective.value()),
        )
        return solve_stats

    def solve_milp(
        self,
        prob: LpProblem,
        warm_start: bool = False,  # noqa: FBT001, FBT002
    ) -> ModelLPMILPStats:
        """Solve problem with CBC.

        Parameters
        ----------
        prob : LpProblem
            Problem to solve
        warm_start : bool, default=False
            Warm start

        Returns
        -------
        MIPStats
            MIP statistics

        Raises
        ------
        NoSolutionError
            If no solution is found
        SolverTimeLimitReached
            If solver time limit is reached

        """
        solve_stats = ModelLPMILPStats(
            cast(UniqueModels, prob.name),
            prob.numVariables(),
            prob.numConstraints(),
        )
        if self._print_log:
            _LOGGER.info("Solve stats:")

        if not warm_start:
            prob.solve(
                PULP_CBC_CMD(
                    mip=False,
                    timeLimit=self._time_limit,
                    msg=self._print_log,
                    warmStart=warm_start,
                ),
            )
            relax_stats = LPStats(
                LpStatus[prob.status],
                prob.solutionTime,
                prob.solutionCpuTime,
            )
            solve_stats.set_relax_stats(relax_stats)
            if prob.status == LpSolutionInfeasible:
                raise NoSolutionError(solve_stats)
            if prob.status == LpStatusNotSolved:
                raise SolverTimeLimitReachedError(solve_stats)
            relax_stats.set_objective_value(
                float(prob.objective.value()),
            )

        prob.solve(
            PULP_CBC_CMD(
                warmStart=True,
                logPath=self.__LOG_FILE_PATH,
                msg=self._print_log,
                timeLimit=self._time_limit,
            ),
        )
        iterations, node_count = self.__get_iterations_node_count()

        milp_stats = WarmMILPStats(
            LpStatus[prob.status],
            prob.solutionTime,
            prob.solutionCpuTime,
            iterations=iterations,
            node_count=node_count,
        )
        solve_stats.set_milp_stats(milp_stats)
        if prob.status == LpSolutionInfeasible:
            raise NoSolutionError(solve_stats)
        if prob.status == LpStatusNotSolved:
            raise SolverTimeLimitReachedError(solve_stats)
        milp_stats.set_objective_value(
            float(prob.objective.value()),
        )
        return solve_stats

    def __get_iterations_node_count(self) -> tuple[int | None, int | None]:
        """Use regex to retrieve these informations."""
        iterations, node_count = None, None
        with self.__LOG_FILE_PATH.open(encoding="utf-8") as file:
            for line in file:
                match = re.search(
                    r"Cbc0001I .+? (\d+) iterations and (\d+) nodes .+?",
                    line,
                )
                if match is not None:
                    iterations, node_count = (
                        int(match.group(1)),
                        int(match.group(2)),
                    )
        self.__LOG_FILE_PATH.unlink()
        return iterations, node_count


class GurobiSolver(Solver):
    """Gurobi solver class."""

    _NAME = "Gurobi"

    def solve_lp(
        self,
        prob: LpProblem,
        warm_start: bool = False,  # noqa: FBT001, FBT002
    ) -> ModelLPStats:
        """Solve LP with CBC LP solver.

        Parameters
        ----------
        prob : LpProblem
            Problem to solve
        warm_start : bool, default=False
            Warm start

        Returns
        -------
        ModelLPStats
            LP statistics

        Raises
        ------
        NoSolutionError
            If no solution is found
        SolverTimeLimitReached
            If solver time limit is reached

        """
        if self._print_log:
            _LOGGER.info("Solve stats:")
        prob.solve(
            GUROBI(
                mip=False,
                timeLimit=self._time_limit,
                msg=self._print_log,
                warmStart=warm_start,
            ),
        )
        lp_stats = LPStats(
            LpStatus[prob.status],
            prob.solutionTime,
            prob.solutionCpuTime,
        )
        solve_stats = ModelLPStats(
            cast(UniqueModels, prob.name),
            prob.numVariables(),
            prob.numConstraints(),
            lp_stats,
        )
        if prob.status == LpSolutionInfeasible:
            raise NoSolutionError(solve_stats)
        if prob.status == LpStatusNotSolved:
            raise SolverTimeLimitReachedError(solve_stats)
        lp_stats.set_objective_value(
            float(prob.objective.value()),
        )
        return solve_stats

    def solve_milp(
        self,
        prob: LpProblem,
        warm_start: bool = False,  # noqa: FBT001, FBT002
    ) -> ModelLPMILPStats:
        """Solve problem with Gurobi.

        Parameters
        ----------
        prob : LpProblem
            Problem to solve
        warm_start : bool, default=False
            Warm start

        Returns
        -------
        MIPStats
            MIP statistics

        Raises
        ------
        NoSolutionError
            If no solution is found
        SolverTimeLimitReached
            If solver time limit is reached

        """
        if self._print_log:
            _LOGGER.info("Solve stats:")
        solve_stats = ModelLPMILPStats(
            cast(UniqueModels, prob.name),
            prob.numVariables(),
            prob.numConstraints(),
        )
        if not warm_start:
            prob.solve(
                GUROBI(
                    mip=False,
                    timeLimit=self._time_limit,
                    msg=self._print_log,
                    warmStart=warm_start,
                ),
            )
            relax_stats = LPStats(
                LpStatus[prob.status],
                prob.solutionTime,
                prob.solutionCpuTime,
            )
            solve_stats.set_relax_stats(relax_stats)
            if prob.status == LpSolutionInfeasible:
                raise NoSolutionError(solve_stats)
            if prob.status == LpStatusNotSolved:
                raise SolverTimeLimitReachedError(solve_stats)
            relax_stats.set_objective_value(
                float(prob.objective.value()),
            )

        prob.solve(
            GUROBI(
                warmStart=True,
                timeLimit=self._time_limit,
                msg=self._print_log,
            ),
        )
        milp_stats = WarmMILPStats(
            LpStatus[prob.status],
            prob.solutionTime,
            prob.solutionCpuTime,
            iterations=prob.solverModel.IterCount,
            node_count=prob.solverModel.NodeCount,
        )
        solve_stats.set_milp_stats(milp_stats)
        if prob.status == LpSolutionInfeasible:
            raise NoSolutionError(solve_stats)
        if prob.status == LpStatusNotSolved:
            raise SolverTimeLimitReachedError(solve_stats)
        milp_stats.set_objective_value(
            float(prob.objective.value()),
        )
        return solve_stats


# ============================================================================ #
#                                   FUNCTIONS                                  #
# ============================================================================ #
def solver_from_dict(solver_config_dict: dict[str, Any]) -> Solver:
    """Convert dict to solver.

    Parameters
    ----------
    solver_config_dict : dict
        Dictionnary of solver config

    Returns
    -------
    Solver
        Solver

    Raises
    ------
    ValueError
        If the solver is unknown

    """
    if solver_config_dict[Solver.KEY_NAME] == GurobiSolver.name():
        return GurobiSolver.from_dict(solver_config_dict)
    if solver_config_dict[Solver.KEY_NAME] == CBCSolver.name():
        return CBCSolver.from_dict(solver_config_dict)
    msg = f"Unknown solver: {solver_config_dict[Solver.KEY_NAME]}"
    raise ValueError(msg)


# ============================================================================ #
#                                  EXCEPTIONS                                  #
# ============================================================================ #
class NoSolutionError(Exception):
    """No solution exception."""

    def __init__(self, solve_stats: ModelSolveStats) -> None:
        """Initialize the exception."""
        super().__init__()
        self.__solve_stats: ModelSolveStats = solve_stats

    def solve_stats(self) -> ModelSolveStats:
        """Solve statistics.

        Returns
        -------
        SolveStats
            Solve statistics

        """
        return self.__solve_stats

    def __str__(self) -> str:
        """Print the exception message.

        Returns
        -------
        str
            Exception message

        """
        return (
            f"The problem is unfeasible, see the solve statistics:\n\n"
            f"{self.__solve_stats}"
        )


class SolverTimeLimitReachedError(Exception):
    """Solver time limit reached exception."""

    def __init__(self, solve_stats: ModelSolveStats) -> None:
        """Initialize the exception."""
        super().__init__()
        self.__solve_stats: ModelSolveStats = solve_stats

    def solve_stats(self) -> ModelSolveStats:
        """Solve statistics.

        Returns
        -------
        SolveStats
            Solve statistics

        """
        return self.__solve_stats

    def __str__(self) -> str:
        """Print the exception message.

        Returns
        -------
        str
            Exception message

        """
        return (
            f"The solver time limit is reached,"
            " see the solve statistics:\n\n"
            f"{self.__solve_stats}"
        )



"""Application module for binary matrix operations."""

from __future__ import annotations

import logging
import sys
from dataclasses import dataclass
from pathlib import Path  # noqa: TCH003
from typing import Callable

import typer

from qbc_dsbm.binary_matrix import (
    BinMatrix,
    BinMatrixStats,
    DegeneratedMatrixError,
    a_inter_b_columns,
    a_inter_b_rows,
    a_minus_b_columns,
)

_LOGGER = logging.getLogger(__name__)
_LOGGER.setLevel(logging.INFO)
console_handler = logging.StreamHandler()
formatter = logging.Formatter(
    "%(asctime)s - %(name)s [%(levelname)s] %(message)s",
)
console_handler.setFormatter(formatter)
_LOGGER.addHandler(console_handler)


@dataclass
class MatricesOperationArgs:
    """Matrices operation args and opts."""

    ARG_SUPER_BM_CSV_PATH = typer.Argument(
        help="Path to the super binary matrix CSV file",
    )
    ARG_SUB_BM_CSV_PATH = typer.Argument(
        help="Path to the sub binary matrix CSV file",
    )
    ARG_RESULT_BM_CSV_PATH = typer.Argument(
        help="Path to the resulting binary matrix CSV file",
    )
    OPT_MUTE_LOGGER = typer.Option(
        default=False,
        help="Mute the logger",
    )


APP = typer.Typer()


@APP.command()
def coldiff(
    super_binmatrix_csv: Path = MatricesOperationArgs.ARG_SUPER_BM_CSV_PATH,
    sub_binmatrix_csv: Path = MatricesOperationArgs.ARG_SUB_BM_CSV_PATH,
    result_binmatrix_csv: Path = MatricesOperationArgs.ARG_RESULT_BM_CSV_PATH,
    mute_logger: bool = MatricesOperationArgs.OPT_MUTE_LOGGER,  # noqa: FBT001
) -> None:
    """Produce the binary matrix column-intersection between two binary matrices.

    It keeps the rows but differs on the columns.

    """
    _LOGGER.info("Beginning col-diff operation")
    if mute_logger:
        _LOGGER.setLevel(logging.CRITICAL)
    super_matrix, sub_matrix = __init_super_sub_bm(
        super_binmatrix_csv,
        sub_binmatrix_csv,
    )
    result_matrix = __execute_op(super_matrix, sub_matrix, a_minus_b_columns)
    __finish_result_matrix(result_binmatrix_csv, result_matrix)


@APP.command()
def colinter(
    super_binmatrix_csv: Path = MatricesOperationArgs.ARG_SUPER_BM_CSV_PATH,
    sub_binmatrix_csv: Path = MatricesOperationArgs.ARG_SUB_BM_CSV_PATH,
    result_binmatrix_csv: Path = MatricesOperationArgs.ARG_RESULT_BM_CSV_PATH,
    mute_logger: bool = MatricesOperationArgs.OPT_MUTE_LOGGER,  # noqa: FBT001
) -> None:
    """Produce the binary matrix column-intersection between two binary matrices.

    It keeps the rows but differs on the columns.

    """
    _LOGGER.info("Beginning col-inter operation")
    if mute_logger:
        _LOGGER.setLevel(logging.CRITICAL)
    super_matrix, sub_matrix = __init_super_sub_bm(
        super_binmatrix_csv,
        sub_binmatrix_csv,
    )
    result_matrix = __execute_op(super_matrix, sub_matrix, a_inter_b_columns)
    __finish_result_matrix(result_binmatrix_csv, result_matrix)


@APP.command()
def rowdiff(
    super_binmatrix_csv: Path = MatricesOperationArgs.ARG_SUPER_BM_CSV_PATH,
    sub_binmatrix_csv: Path = MatricesOperationArgs.ARG_SUB_BM_CSV_PATH,
    result_binmatrix_csv: Path = MatricesOperationArgs.ARG_RESULT_BM_CSV_PATH,
    mute_logger: bool = MatricesOperationArgs.OPT_MUTE_LOGGER,  # noqa: FBT001
) -> None:
    """Produce the binary matrix row-difference between two binary matrices.

    It keeps the columns but differs on the rows.

    """
    _LOGGER.info("Beginning row-diff operation")
    if mute_logger:
        _LOGGER.setLevel(logging.CRITICAL)
    super_matrix, sub_matrix = __init_super_sub_bm(
        super_binmatrix_csv,
        sub_binmatrix_csv,
    )
    result_matrix = __execute_op(super_matrix, sub_matrix, a_inter_b_rows)
    __finish_result_matrix(result_binmatrix_csv, result_matrix)


@APP.command()
def rowinter(
    super_binmatrix_csv: Path = MatricesOperationArgs.ARG_SUPER_BM_CSV_PATH,
    sub_binmatrix_csv: Path = MatricesOperationArgs.ARG_SUB_BM_CSV_PATH,
    result_binmatrix_csv: Path = MatricesOperationArgs.ARG_RESULT_BM_CSV_PATH,
    mute_logger: bool = MatricesOperationArgs.OPT_MUTE_LOGGER,  # noqa: FBT001
) -> None:
    """Produce the binary matrix row-intersection between two binary matrices.

    It keeps the columns but differs on the rows.

    """
    _LOGGER.info("Beginning row-inter operation")
    if mute_logger:
        _LOGGER.setLevel(logging.CRITICAL)
    super_matrix, sub_matrix = __init_super_sub_bm(
        super_binmatrix_csv,
        sub_binmatrix_csv,
    )
    result_matrix = __execute_op(super_matrix, sub_matrix, a_inter_b_rows)
    __finish_result_matrix(result_binmatrix_csv, result_matrix)


def __init_super_sub_bm(
    super_binmatrix_csv: Path,
    sub_binmatrix_csv: Path,
) -> tuple[BinMatrix, BinMatrix]:
    super_matrix = BinMatrix.from_csv(super_binmatrix_csv)
    sub_matrix = BinMatrix.from_csv(sub_binmatrix_csv)
    _LOGGER.info(
        "Super matrix:\n%s\n",
        BinMatrixStats.from_bin_matrix(super_matrix),
    )
    _LOGGER.info(
        "Sub matrix:\n%s\n",
        BinMatrixStats.from_bin_matrix(sub_matrix),
    )
    return super_matrix, sub_matrix


def __execute_op(
    super_matrix: BinMatrix,
    sub_matrix: BinMatrix,
    op_func: Callable[[BinMatrix, BinMatrix], BinMatrix],
) -> BinMatrix:
    try:
        return op_func(super_matrix, sub_matrix)
    except DegeneratedMatrixError as error:
        _LOGGER.critical("%s", error)
        sys.exit(1)


def __finish_result_matrix(
    result_binmatrix_csv: Path,
    result_matrix: BinMatrix,
) -> None:
    _LOGGER.info(
        "Resulting matrix:\n%s\n",
        BinMatrixStats.from_bin_matrix(result_matrix),
    )
    result_matrix.to_csv(result_binmatrix_csv)



"""Binary matrix create module."""

from __future__ import annotations

import logging
from typing import TYPE_CHECKING

import numpy as np
import pandas as pd
from sklearn.impute import KNNImputer

from qbc_dsbm.binary_matrix.items import BinMatrix

if TYPE_CHECKING:
    from qbc_dsbm.binary_matrix import configs

_LOGGER = logging.getLogger(__name__)


def generate_bin_dataframe(
    number_of_rows: int,
    number_of_columns: int,
    sparsicity: float,
) -> pd.DataFrame:
    """Generate a binary dataframe.

    Parameters
    ----------
    number_of_rows : int
        Number of rows
    number_of_columns : int
        Number of columns
    sparsicity : float
        Sparsicity of the matrix

    Returns
    -------
    pd.DataFrame
        A binary dataframe

    """
    data = np.random.default_rng().binomial(
        1,
        p=(1 - sparsicity),
        size=(number_of_rows, number_of_columns),
    )
    return pd.DataFrame(
        data,
        index=pd.Index(range(number_of_rows), name="row"),
        columns=pd.Index(range(number_of_columns), name="column"),
    )


def encapsulate_permutate_bin_dataframe(
    number_of_rows: int,
    number_of_columns: int,
    sparsicity: float,
    number_of_rows_super: int,
    number_of_columns_super: int,
) -> tuple[BinMatrix, BinMatrix]:
    """Create a binary dataframe containing a hidden binary dataframe.

    There exists a permutation of the rows and the columns of the super
    binary matrix such its top-left corner contains the smaller binary matrix.

    Parameters
    ----------
    number_of_rows : int
        Number of rows of the contained matrix
    number_of_columns : int
        Number of columns of the contained matrix
    sparsicity : float
        Sparsicity of the contained binary matrix
    number_of_rows_super : int
        Number of rows of the containing matrix >= number_of_rows
    number_of_columns_super : int
        Number of columns of the containing matrix >= number_of_columns

    Returns
    -------
    BinMatrix
        Contained binary matrix
    BinMatrix
        Containing binary matrix

    Raises
    ------
    ValueError
        If number_of_rows_super < number_of_rows
        or if number_of_columns_super < number_of_columns

    """
    if number_of_rows_super < number_of_rows:
        msg = (
            "super number of rows < sub number of rows"
            f" ({number_of_rows_super} < {number_of_rows})"
        )
        raise ValueError(msg)
    if number_of_columns_super < number_of_columns:
        msg = (
            "super number of columns < sub number of columns"
            f" ({number_of_columns_super} < {number_of_columns})"
        )
        raise ValueError(msg)

    contained_dataframe = generate_bin_dataframe(
        number_of_rows,
        number_of_columns,
        sparsicity,
    )
    super_dataframe = pd.DataFrame(
        False,  # noqa: FBT003
        index=range(number_of_rows_super),
        columns=range(number_of_columns_super),
    )
    super_dataframe.update(contained_dataframe)
    return (
        BinMatrix.new_unsafe(contained_dataframe),
        BinMatrix.new_unsafe(
            super_dataframe.reindex(
                index=np.random.default_rng().permutation(super_dataframe.index),
                columns=np.random.default_rng().permutation(super_dataframe.columns),
            ),
        ),
    )


def generate_bin_matrix(
    number_of_rows: int,
    number_of_columns: int,
    sparsity: float,
) -> BinMatrix:
    """Generate a binary matrix.

    Parameters
    ----------
    number_of_rows : int
        Number of rows
    number_of_columns : int
        Number of columns
    sparsity : float
        Sparsity of the matrix

    Returns
    -------
    BinMatrix
        A binary matrix

    """
    return BinMatrix.new_unsafe(
        generate_bin_dataframe(number_of_rows, number_of_columns, sparsity),
    )


def complete_dataframe_with_knn(
    unfilled_df: pd.DataFrame,
    config: configs.KNNCompletetionConfig,
) -> BinMatrix:
    """Complete a dataframe with KNN.

    Parameters
    ----------
    unfilled_df : pd.DataFrame
        Dataframe with unfilled values
    config : bm_cfg.KNNCompletetionConfig
        KNN completion parameters

    Returns
    -------
    BinMatrix
        KNN-completed binary matrix

    """
    _LOGGER.info("Complete the dataframe with KNN")
    _LOGGER.info("Configuration:\n%s\n", config)

    m, n = unfilled_df.shape
    if m > 1 and n > 1:
        imputer = KNNImputer(n_neighbors=config.number_of_neighbours())
        matrix = imputer.fit_transform(unfilled_df)
        matrix[(matrix >= config.threshold())] = 1
        matrix[(matrix < config.threshold())] = 0
        matrix = pd.DataFrame(
            matrix,
            index=unfilled_df.index,
            columns=unfilled_df.columns,
        )
    else:
        matrix = unfilled_df

    return BinMatrix(matrix)



"""Input-output configurations."""

from __future__ import annotations

from pathlib import Path
from typing import Any

from qbc_dsbm import (
    QBCExactModels,
    QBCHeuristicModels,
    UniqueModels,
    string_to_model_enum,
)
from qbc_dsbm.models.solvers import Solver, solver_from_dict


# ============================================================================ #
#                                     CLASS                                    #
# ============================================================================ #
class IOConfig:
    """Input-output config class.

    Parameters
    ----------
    csv_path : Path
        CSV path
    model : UniqueModels
        Model
    epsilon : float
        Epsilon
    solver : Solver
        Solver

    """

    __KEY_CSV_PATH: str = "csv_path"
    __KEY_MODEL: str = "model"
    __KEY_EPSILON: str = "epsilon"
    __KEY_SOLVER: str = "solver"
    __KEY_MIN_NUMBER_OF_ROWS: str = "min_number_of_rows"
    __KEY_MIN_NUMBER_OF_COLUMNS: str = "min_number_of_columns"
    __KEY_CHECK_SATISFIABILITY: str = "check_satisfiability"

    @classmethod
    def from_dict(cls, io_config_dict: dict[str, Any]) -> IOConfig:
        """Convert dict to IOConfig.

        Parameters
        ----------
        io_config_dict : dict
            Dictionnary of IO config

        Returns
        -------
        IOConfig
            IOConfig

        """
        return cls(
            Path(io_config_dict[cls.__KEY_CSV_PATH]),
            string_to_model_enum(io_config_dict[cls.__KEY_MODEL]),
            io_config_dict[cls.__KEY_EPSILON],
            solver_from_dict(io_config_dict[cls.__KEY_SOLVER]),
            io_config_dict[cls.__KEY_MIN_NUMBER_OF_ROWS],
            io_config_dict[cls.__KEY_MIN_NUMBER_OF_COLUMNS],
            io_config_dict[cls.__KEY_CHECK_SATISFIABILITY],
        )

    def __init__(  # noqa: PLR0913
        self,
        csv_path: Path,
        model: UniqueModels,
        epsilon: float,
        solver: Solver,
        min_number_of_rows: int,
        min_number_of_columns: int,
        check_satisfiability: bool,  # noqa: FBT001
    ) -> None:
        """Initialize the configuration."""
        self.__csv_path: Path = csv_path
        self.__model: UniqueModels = model
        self.__epsilon: float = epsilon
        self.__solver: Solver = solver
        self.__min_number_of_rows: int = min_number_of_rows
        self.__min_number_of_columns: int = min_number_of_columns
        self.__check_satisfiability: bool = check_satisfiability

    def csv_path(self) -> Path:
        """CSV file path.

        Returns
        -------
        Path
            CSV file path

        """
        return self.__csv_path

    def model(self) -> UniqueModels:
        """Model string name.

        Returns
        -------
        UniqueModels
            Model string name

        """
        return self.__model

    def epsilon(self) -> float:
        """Epsilon.

        Returns
        -------
        float
            Epsilon

        """
        return self.__epsilon

    def solver(self) -> Solver:
        """Solver.

        Returns
        -------
        Solver
            Solver

        """
        return self.__solver

    def min_number_of_rows(self) -> int:
        """Minimum number of rows.

        Returns
        -------
        int
            Minimum number of rows

        """
        return self.__min_number_of_rows

    def min_number_of_columns(self) -> int:
        """Minimum number of columns.

        Returns
        -------
        int
            Minimum number of columns

        """
        return self.__min_number_of_columns

    def check_satisfiability(self) -> bool:
        """Check satisfiability.

        Returns
        -------
        bool
            Check satisfiability

        """
        return self.__check_satisfiability

    def to_dict(self) -> dict[str, Any]:
        """Convert to dict.

        Returns
        -------
        dict[str, Any]
            Dictionary

        """
        return {
            self.__KEY_CSV_PATH: str(self.__csv_path),
            self.__KEY_MODEL: self.__model.value,
            self.__KEY_EPSILON: self.__epsilon,
            self.__KEY_SOLVER: self.__solver.to_dict(),
            self.__KEY_MIN_NUMBER_OF_ROWS: self.__min_number_of_rows,
            self.__KEY_MIN_NUMBER_OF_COLUMNS: self.__min_number_of_columns,
            self.__KEY_CHECK_SATISFIABILITY: self.__check_satisfiability,
        }

    def __str__(self) -> str:
        """Print the IOConfig parameters."""
        return "\n".join(
            [
                f"CSV path: {self.__csv_path}",
                f"Model name: {self.__model.value}",
                f"Epsilon: {self.__epsilon}",
                f"Solver: {self.__solver.name()}",
                f"Minimum number of rows: {self.__min_number_of_rows}",
                f"Minimum number of columns: {self.__min_number_of_columns}",
                f"Check satisfiability: {self.__check_satisfiability}",
            ],
        )


class IOConfigDivCqr:
    """Input-output config class for divide and conquer method.

    Parameters
    ----------
    csv_path : Path
        The path to the csv file
    heuristic_model : QBCHeuristicModelStringName
        The model to use
    heuristic_epsilon : float
        Percentage of errors (remaining zeros) in the result submatrix
    exact_model : QBCExactModelStringName
        The model to use
    exact_epsilon : float
        Percentage of errors (remaining zeros) in the result submatrix
    solver : Solver
        The solver to use

    """

    __KEY_CSV_PATH: str = "csv_path"
    __KEY_HEURISTIC_MODEL: str = "heuristic_model"
    __KEY_HEURISTIC_EPSILON: str = "heuristic_epsilon"
    __KEY_EXACT_MODEL: str = "exact_model"
    __KEY_EXACT_EPSILON: str = "exact_epsilon"
    __KEY_SOLVER: str = "solver"

    @classmethod
    def from_dict(cls, io_config_dict: dict[str, Any]) -> IOConfigDivCqr:
        """Instantiate IOConfigDivCqr from dict.

        Parameters
        ----------
        io_config_dict : dict[str, Any]
            Dictionary containing the IOConfigDivCqr parameters

        Returns
        -------
        IOConfigDivCqr
            IOConfigDivCqr

        """
        return cls(
            Path(io_config_dict[cls.__KEY_CSV_PATH]),
            QBCHeuristicModels(io_config_dict[cls.__KEY_HEURISTIC_MODEL]),
            io_config_dict[cls.__KEY_HEURISTIC_EPSILON],
            QBCExactModels(io_config_dict[cls.__KEY_EXACT_MODEL]),
            io_config_dict[cls.__KEY_EXACT_EPSILON],
            solver_from_dict(io_config_dict[cls.__KEY_SOLVER]),
        )

    def __init__(  # noqa: PLR0913
        self,
        csv_path: Path,
        heuristic_model: QBCHeuristicModels,
        heuristic_epsilon: float,
        exact_model: QBCExactModels,
        exact_epsilon: float,
        solver: Solver,
    ) -> None:
        """Initialize the configuration."""
        self.__csv_path: Path = csv_path
        self.__heuristic_model: QBCHeuristicModels = heuristic_model
        self.__heuristic_epsilon: float = heuristic_epsilon
        self.__exact_model: QBCExactModels = exact_model
        self.__exact_epsilon: float = exact_epsilon
        self.__solver: Solver = solver

    def csv_path(self) -> Path:
        """CSV file path.

        Returns
        -------
        Path
            CSV file path

        """
        return self.__csv_path

    def heuristic_model(self) -> QBCHeuristicModels:
        """Heuristic model.

        Returns
        -------
        QBCHeuristicModels
            Heuristic model

        """
        return self.__heuristic_model

    def heuristic_epsilon(self) -> float:
        """Epsilon for the heuristic model.

        Returns
        -------
        float
            Epsilon

        """
        return self.__heuristic_epsilon

    def exact_model(self) -> QBCExactModels:
        """Exact model.

        Returns
        -------
        QBCExactModels
            Exact model

        """
        return self.__exact_model

    def exact_epsilon(self) -> float:
        """Epsilon for the exact model.

        Returns
        -------
        float
            Epsilon

        """
        return self.__exact_epsilon

    def solver(self) -> Solver:
        """Solver.

        Returns
        -------
        Solver
            Solver

        """
        return self.__solver

    def to_dict(self) -> dict[str, Any]:
        """Convert to dict.

        Returns
        -------
        dict[str, Any]
            Dictionary

        """
        return {
            self.__KEY_CSV_PATH: str(self.__csv_path),
            self.__KEY_HEURISTIC_MODEL: self.__heuristic_model.value,
            self.__KEY_HEURISTIC_EPSILON: self.__heuristic_epsilon,
            self.__KEY_EXACT_MODEL: self.__exact_model.value,
            self.__KEY_EXACT_EPSILON: self.__exact_epsilon,
            self.__KEY_SOLVER: self.__solver.to_dict(),
        }

    def __str__(self) -> str:
        """Print the IOConfigDAC parameters."""
        return "\n".join(
            [
                "----------",
                "Parameters",
                "----------",
                "",
                f"CSV path: {self.__csv_path}",
                f"Heuristic model name: {self.__heuristic_model.value}",
                f"Heuristic epsilon: {self.__heuristic_epsilon}",
                f"Exact model name: {self.__exact_model.name}",
                f"Exact epsilon: {self.__exact_epsilon}",
                f"Solver: {self.__solver.name()}",
            ],
        )



"""MIP stats module."""

from __future__ import annotations

from typing import TYPE_CHECKING, Any

if TYPE_CHECKING:
    from qbc_dsbm import UniqueModels

import enum


class ModelType(enum.Enum):
    """Model type enum."""

    LP = "LP"
    """LP model."""

    MILP = "MILP"
    """MILP model."""


# ============================================================================ #
#                                    CLASSES                                   #
# ============================================================================ #
class LPStats:
    """LP statistics."""

    __STATUS = "status"
    __USER_TIME = "user_time"
    __CPU_TIME = "cpu_time"
    __OBJ_VALUE = "objective_value"

    @classmethod
    def from_dict(cls, relax_stats_dict: dict[str, Any]) -> LPStats:
        """Convert dict to RelaxStats.

        Parameters
        ----------
        relax_stats_dict : dict
            Dictionnary of relax stats

        Returns
        -------
        RelaxStats
            RelaxStats

        """
        return cls(
            status=relax_stats_dict[cls.__STATUS],
            relax_user_time=relax_stats_dict[cls.__USER_TIME],
            relax_cpu_time=relax_stats_dict[cls.__CPU_TIME],
            objective_value=relax_stats_dict[cls.__OBJ_VALUE],
        )

    def __init__(
        self,
        status: str,
        relax_user_time: float,
        relax_cpu_time: float,
        objective_value: float | None = None,
    ) -> None:
        """Initialize the stats."""
        self.__status = status
        self.__user_time = relax_user_time
        self.__cpu_time = relax_cpu_time
        self.__objective_value = objective_value

    def status(self) -> str:
        """Status.

        Returns
        -------
        str
            Status

        """
        return self.__status

    def user_time(self) -> float:
        """User time (in seconds).

        Returns
        -------
        float
            User time

        """
        return self.__user_time

    def cpu_time(self) -> float:
        """CPU time (in seconds).

        Returns
        -------
        float
            CPU time

        """
        return self.__cpu_time

    def objective_value(self) -> float | None:
        """Objective value.

        Returns
        -------
        float or None
            Objective value

        """
        return self.__objective_value

    def set_objective_value(self, value: float) -> None:
        """Set objective value.

        Parameters
        ----------
        value : float
            Objective value

        """
        self.__objective_value = value

    def to_dict(self) -> dict[str, Any]:
        """Convert to dict.

        Returns
        -------
        dict
            Dict

        """
        return {
            self.__STATUS: self.__status,
            self.__USER_TIME: self.__user_time,
            self.__CPU_TIME: self.__cpu_time,
            self.__OBJ_VALUE: self.__objective_value,
        }

    def __str__(self) -> str:
        """Print the statistics.

        Returns
        -------
        str
            Statistics

        """
        return "\n".join(
            [
                f"Status: {self.__status}",
                f"User time: {self.__user_time}",
                f"CPU time: {self.__cpu_time}",
                f"Objective value: {self.__objective_value}",
            ],
        )


class WarmMILPStats:
    """Statistics for MIP problems."""

    __STATUS = "status"
    __USER_TIME = "user_time"
    __CPU_TIME = "cpu_time"
    __OBJ_VALUE = "objective_value"
    __ITERATIONS = "iterations"
    __NODE_COUNT = "node_count"

    @classmethod
    def from_dict(cls, mip_stats_dict: dict[str, Any]) -> WarmMILPStats:
        """Convert dict to MIPStats.

        Parameters
        ----------
        mip_stats_dict : dict
            Dictionnary of MIP stats

        Returns
        -------
        MIPStats
            MIPStats

        """
        return WarmMILPStats(
            status=mip_stats_dict[cls.__STATUS],
            user_time=mip_stats_dict[cls.__USER_TIME],
            cpu_time=mip_stats_dict[cls.__CPU_TIME],
            obj_value=mip_stats_dict[cls.__OBJ_VALUE],
            iterations=mip_stats_dict[cls.__ITERATIONS],
            node_count=mip_stats_dict[cls.__NODE_COUNT],
        )

    def __init__(  # noqa: PLR0913
        self,
        status: str,
        user_time: float,
        cpu_time: float,
        obj_value: float | None = None,
        iterations: int | None = None,
        node_count: int | None = None,
    ) -> None:
        """Initialize the stats."""
        self.__status: str = status
        self.__user_time: float = user_time
        self.__cpu_time: float = cpu_time
        self.__obj_value: float | None = obj_value
        self.__iterations: int | None = iterations
        self.__node_count: int | None = node_count

    def status(self) -> str:
        """Solver status.

        Returns
        -------
        str
            Solver status

        """
        return self.__status

    def user_time(self) -> float:
        """User time (in seconds).

        Returns
        -------
        float
            User time

        """
        return self.__user_time

    def cpu_time(self) -> float:
        """CPU time (in seconds).

        Returns
        -------
        float
            CPU time

        """
        return self.__cpu_time

    def objective_value(self) -> float | None:
        """Objective value.

        Returns
        -------
        float or None
            Objective value

        """
        return self.__obj_value

    def iterations(self) -> int | None:
        """Give the number of iterations used to solve the MIP.

        Returns
        -------
        int
            Number of iterations

        """
        return self.__iterations

    def node_count(self) -> int | None:
        """Give the number of nodes explored by the solver.

        Returns
        -------
        int or None
            Number of nodes

        """
        return self.__node_count

    def set_objective_value(self, value: float) -> None:
        """Set objective value.

        Parameters
        ----------
        value : float
            Objective value

        """
        self.__obj_value = value

    def set_iterations(self, value: int) -> None:
        """Set number of iterations.

        Parameters
        ----------
        value : int
            Number of iterations

        """
        self.__iterations = value

    def set_node_count(self, value: int) -> None:
        """Set number of nodes.

        Parameters
        ----------
        value : int
            Number of nodes

        """
        self.__node_count = value

    def to_dict(self) -> dict:
        """Convert MIPStats to dict.

        Returns
        -------
        dict
            Dictionnary

        """
        return {
            self.__STATUS: self.__status,
            self.__USER_TIME: self.__user_time,
            self.__CPU_TIME: self.__cpu_time,
            self.__OBJ_VALUE: self.__obj_value,
            self.__ITERATIONS: self.__iterations,
            self.__NODE_COUNT: self.__node_count,
        }

    def __str__(self) -> str:
        """Give the string representation.

        Returns
        -------
        str
            String representation

        """
        return "\n".join(
            [
                f"Status: {self.__status}",
                f"User time: {self.__user_time}",
                f"CPU time: {self.__cpu_time}",
                f"Objective value: {self.__obj_value}",
                f"Iterations: {self.__iterations}",
                f"Node count: {self.__node_count}",
            ],
        )


class ModelSolveStats:
    """Solve stats parent class."""

    _MODEL_TYPE = ModelType.MILP

    _MODEL_TYPE_KEY = "model_type"
    _MODEL_NAME_KEY = "model"

    _NUMBER_VARIABLES_KEY = "number_variables"
    _NUMBER_CONSTRAINTS_KEY = "number_constraints"

    @classmethod
    def model_type(cls) -> ModelType:
        """Model type.

        Returns
        -------
        ModelType
            Model type

        """
        return cls._MODEL_TYPE

    @classmethod
    def model_type_from_dict(cls, stats_dict: dict) -> ModelType:
        """Retrun model type from a dictionnary.

        Parameters
        ----------
        stats_dict : dict
            Dictionary

        Returns
        -------
        ModelType
            Model type

        Raises
        ------
        ValueError
            Model type not found

        """
        if cls._MODEL_TYPE_KEY in stats_dict:
            return ModelType(stats_dict[cls._MODEL_TYPE_KEY])
        msg = f"Model type not found in {stats_dict}."
        raise ValueError(msg)

    @classmethod
    def from_dict(cls, stats_dict: dict) -> ModelSolveStats:
        """Convert dict to ModelSolveStats.

        Parameters
        ----------
        stats_dict : dict
            Dictionary

        Returns
        -------
        ModelSolveStats
            ModelSolveStats

        """
        return cls(
            stats_dict[cls._MODEL_NAME_KEY],
            stats_dict[cls._NUMBER_VARIABLES_KEY],
            stats_dict[cls._NUMBER_CONSTRAINTS_KEY],
        )

    def __init__(
        self,
        model: UniqueModels,
        number_variables: int,
        number_constraints: int,
    ) -> None:
        """Initialize the stats."""
        self._model = model
        self._number_variables = number_variables
        self._number_constraints = number_constraints

    def model(self) -> UniqueModels:
        """Model.

        Returns
        -------
        UniqueModels
            Model

        """
        return self._model

    def number_variables(self) -> int:
        """Give the number of variables in the MIP.

        Returns
        -------
        int
            Number of variables

        """
        return self._number_variables

    def number_constraints(self) -> int:
        """Give the number of constraints in the MIP.

        Returns
        -------
        int
            Number of constraints

        """
        return self._number_constraints

    def to_dict(self) -> dict:
        """Convert SolveStats to dict.

        Returns
        -------
        dict
            Dictionnary

        """
        return {
            self._MODEL_TYPE_KEY: self._MODEL_TYPE.value,
            self._MODEL_NAME_KEY: self._model,
            self._NUMBER_VARIABLES_KEY: self._number_variables,
            self._NUMBER_CONSTRAINTS_KEY: self._number_constraints,
        }

    def __str__(self) -> str:
        """Give the string representation.

        Returns
        -------
        str
            String representation

        """
        return "\n".join(
            [
                f"Model type: {self._MODEL_TYPE.value}",
                f"Model: {self._model}",
                f"Number of variables: {self._number_variables}",
                f"Number of constraints: {self._number_constraints}",
            ],
        )


class ModelLPStats(ModelSolveStats):
    """LP solve stats."""

    _MODEL_TYPE = ModelType.LP

    __SOLVE_STATS_KEY = "solve_stats"

    @classmethod
    def from_dict(cls, stats_dict: dict) -> ModelLPStats:
        """Convert dict to ModelLPStats.

        Parameters
        ----------
        stats_dict : dict
            Dictionary

        Returns
        -------
        ModelLPStats
            ModelLPStats

        """
        return cls(
            stats_dict[cls._MODEL_NAME_KEY],
            stats_dict[cls._NUMBER_VARIABLES_KEY],
            stats_dict[cls._NUMBER_CONSTRAINTS_KEY],
            LPStats.from_dict(stats_dict[cls.__SOLVE_STATS_KEY]),
        )

    def __init__(
        self,
        model: UniqueModels,
        number_variables: int,
        number_constraints: int,
        relax_stats: LPStats,
    ) -> None:
        """Initialize the stats."""
        super().__init__(model, number_variables, number_constraints)
        self._relax_stats = relax_stats

    def relax_stats(self) -> LPStats:
        """Relaxation stats.

        Returns
        -------
        LPStats
            Relaxation stats

        """
        return self._relax_stats

    def to_dict(self) -> dict:
        """Convert SolveStats to dict.

        Returns
        -------
        dict
            Dictionnary

        """
        return {
            **super().to_dict(),
            self.__SOLVE_STATS_KEY: self._relax_stats.to_dict(),
        }

    def __str__(self) -> str:
        """Give the string representation.

        Returns
        -------
        str
            String representation

        """
        return "\n".join(
            [
                super().__str__(),
                self._relax_stats.__str__(),
            ],
        )


class ModelLPMILPStats(ModelSolveStats):
    """LP + warm MILP solve stats."""

    _MODEL_TYPE = ModelType.MILP

    __RELAX_STATS_KEY = "relaxation"
    __MILP_STATS_KEY = "milp"

    @classmethod
    def from_dict(cls, stats_dict: dict) -> ModelLPMILPStats:
        """Convert dict to ModelLPMILPStats.

        Parameters
        ----------
        stats_dict : dict
            Dictionary

        Returns
        -------
        ModelLPMILPStats
            ModelLPMILPStats

        """
        return cls(
            stats_dict[cls._MODEL_NAME_KEY],
            stats_dict[cls._NUMBER_VARIABLES_KEY],
            stats_dict[cls._NUMBER_CONSTRAINTS_KEY],
            (
                LPStats.from_dict(stats_dict[cls.__RELAX_STATS_KEY])
                if stats_dict[cls.__RELAX_STATS_KEY] is not None
                else None
            ),
            (
                WarmMILPStats.from_dict(stats_dict[cls.__MILP_STATS_KEY])
                if stats_dict[cls.__MILP_STATS_KEY] is not None
                else None
            ),
        )

    def __init__(
        self,
        model: UniqueModels,
        number_variables: int,
        number_constraints: int,
        relax_stats: LPStats | None = None,
        milp_stats: WarmMILPStats | None = None,
    ) -> None:
        """Initialize the stats."""
        super().__init__(model, number_variables, number_constraints)
        self.__relax_stats: LPStats | None = relax_stats
        self.__milp_stats: WarmMILPStats | None = milp_stats

    def relax_stats(self) -> LPStats | None:
        """Relaxation statistics.

        Returns
        -------
        RelaxStats
            Relaxation statistics

        """
        return self.__relax_stats

    def milp_stats(self) -> WarmMILPStats | None:
        """MILP statistics.

        Returns
        -------
        MILPStats
            MILP statistics

        """
        return self.__milp_stats

    def set_relax_stats(self, relax_stats: LPStats) -> None:
        """Set relaxation statistics.

        Parameters
        ----------
        relax_stats : RelaxStats
            Relaxation statistics

        """
        self.__relax_stats = relax_stats

    def set_milp_stats(self, milp_stats: WarmMILPStats) -> None:
        """Set MILP statistics.

        Parameters
        ----------
        milp_stats : MILPStats
            MILP statistics

        """
        self.__milp_stats = milp_stats

    def to_dict(self) -> dict:
        """Convert ModelLPMILPStats to dict.

        Returns
        -------
        dict
            Dictionnary

        """
        return {
            **super().to_dict(),
            self.__RELAX_STATS_KEY: (
                self.__relax_stats.to_dict() if self.__relax_stats is not None else None
            ),
            self.__MILP_STATS_KEY: (
                self.__milp_stats.to_dict() if self.__milp_stats is not None else None
            ),
        }

    def __str__(self) -> str:
        """Give the string representation.

        Returns
        -------
        str
            String representation

        """
        return "\n".join(
            [
                f"{super().__str__()}",
                "",
                "Relaxation statistics",
                "---------------------",
                f"{self.__relax_stats}",
                "",
                "MILP statistics",
                "---------------",
                f"{self.__milp_stats}",
            ],
        )



"""ILP for the exact densest sub-binary matrix (DSBM) problem."""

from __future__ import annotations

import itertools
from typing import TYPE_CHECKING

from pulp import LpInteger, LpMaximize, LpProblem, LpVariable, lpSum

from qbc_dsbm import QBCExactModels

if TYPE_CHECKING:
    from qbc_dsbm.binary_matrix import BinMatrix

# ============================================================================ #
#                                     TYPES                                    #
# ============================================================================ #
RowColumnChoicesLP = list[LpVariable]
CellChoicesLP = list[list[LpVariable]]


# ============================================================================ #
#                     LP MODEL - MAXIMIZE DESIRED CELL                         #
# ============================================================================ #
def max_ones(
    bin_matrix: BinMatrix,
    epsilon: float,
) -> tuple[LpProblem, RowColumnChoicesLP, RowColumnChoicesLP, CellChoicesLP]:
    """Maximize the number of ones in the matrix.

    Parameters
    ----------
    bin_matrix : BinMatrix
        Binary matrix
    epsilon : float
        Error rate

    Returns
    -------
    LpProblem
        ILP model
    RowColumnChoicesLP
        Row choices
    RowColumnChoicesLP
        Column choices
    CellChoicesLP
        Cell choices

    """
    prob = LpProblem(name=QBCExactModels.MAX_ONES.value, sense=LpMaximize)

    #
    # Variables
    #
    (
        row_choices,
        column_choices,
        cell_choices,
    ) = __row_column_cell_choices_variables(bin_matrix)

    #
    # Objective function
    #
    prob += (
        lpSum(
            bin_matrix[u, v] * cell_choices[u][v]
            for u in range(bin_matrix.number_of_rows())
            for v in range(bin_matrix.number_of_columns())
        ),
        "maximize_ones_in_matrix",
    )

    #
    # Constraints
    #
    __select_cell_select_its_row_column(
        bin_matrix,
        prob,
        row_choices,
        column_choices,
        cell_choices,
    )
    __select_no_more_epsilon_zeros(prob, bin_matrix, epsilon, cell_choices)

    return prob, row_choices, column_choices, cell_choices


# TODO add these options to the other models
def max_ones_compact(
    bin_matrix: BinMatrix,
    epsilon: float,
    min_number_of_rows: int = 0,
    min_number_of_columns: int = 0,
    warm_solution: BinMatrix | None = None,
    check_satisfiability: bool = True,  # noqa: FBT001, FBT002 # TODO use this option to the parent functions
) -> tuple[LpProblem, RowColumnChoicesLP, RowColumnChoicesLP, CellChoicesLP]:
    """Maximize the number of ones in the matrix (compact version).

    Parameters
    ----------
    bin_matrix : BinMatrix
        Binary matrix
    epsilon : float
        Error rate
    min_number_of_rows : int
        Minimum number of rows
    min_number_of_columns : int
        Minimum number of columns
    warm_solution : BinMatrix
        Warm solution
    check_satisfiability : bool
        Check satisfiability, do not solve an optimization problem

    Returns
    -------
    LpProblem
        ILP model
    RowColumnChoicesLP
        Row choices
    RowColumnChoicesLP
        Column choices
    CellChoicesLP
        Cell choices

    """
    prob = LpProblem(name=QBCExactModels.MAX_ONES_COMPACT.value, sense=LpMaximize)

    #
    # Variables
    #
    (
        row_choices,
        column_choices,
        cell_choices,
    ) = __row_column_cell_choices_variables(bin_matrix)

    #
    # Objective function
    #
    if check_satisfiability:
        prob += 0, "dummy_objective"
    else:
        prob += (
            lpSum(
                bin_matrix[u, v] * cell_choices[u][v]
                for u in range(bin_matrix.number_of_rows())
                for v in range(bin_matrix.number_of_columns())
            ),
            "maximize_ones_in_matrix",
        )

    #
    # Constraints
    #
    __select_cell_select_its_row_column_compact(
        bin_matrix,
        prob,
        row_choices,
        column_choices,
        cell_choices,
    )
    __select_no_more_epsilon_zeros(prob, bin_matrix, epsilon, cell_choices)
    __respect_the_minimum_dimensions(
        prob,
        min_number_of_rows,
        min_number_of_columns,
        row_choices,
        column_choices,
    )

    if warm_solution is not None:
        __set_initial_values_from_warm_solution(
            bin_matrix,
            warm_solution,
            row_choices,
            column_choices,
            cell_choices,
        )

    return prob, row_choices, column_choices, cell_choices


def max_surface(
    bin_matrix: BinMatrix,
    epsilon: float,
) -> tuple[LpProblem, RowColumnChoicesLP, RowColumnChoicesLP, CellChoicesLP]:
    """Maximize the surface of the matrix.

    Parameters
    ----------
    bin_matrix : BinMatrix
        Binary matrix
    epsilon : float
        Error rate

    Returns
    -------
    LpProblem
        ILP model
    RowColumnChoicesLP
        Row choices
    RowColumnChoicesLP
        Column choices
    CellChoicesLP
        Cell choices

    """
    prob = LpProblem(name=QBCExactModels.MAX_SURFACE.value, sense=LpMaximize)
    #
    # Variables
    #
    (
        row_choices,
        column_choices,
        cell_choices,
    ) = __row_column_cell_choices_variables(bin_matrix)
    #
    # Objective function
    #
    prob += (
        lpSum(
            cell_choices[u][v]
            for u in range(bin_matrix.number_of_rows())
            for v in range(bin_matrix.number_of_columns())
        ),
        "maximize_surface_matrix",
    )
    #
    # Constraints
    #
    __select_cell_select_its_row_column(
        bin_matrix,
        prob,
        row_choices,
        column_choices,
        cell_choices,
    )
    __select_no_more_epsilon_zeros(prob, bin_matrix, epsilon, cell_choices)

    return prob, row_choices, column_choices, cell_choices


def max_perimeter(
    bin_matrix: BinMatrix,
    epsilon: float,
) -> tuple[LpProblem, RowColumnChoicesLP, RowColumnChoicesLP, CellChoicesLP]:
    """Maximize the perimeter of the matrix.

    Parameters
    ----------
    bin_matrix : BinMatrix
        Binary matrix
    epsilon : float
        Error rate

    Returns
    -------
    LpProblem
        ILP model
    RowColumnChoicesLP
        Row choices
    RowColumnChoicesLP
        Column choices
    CellChoicesLP
        Cell choices

    """
    prob = LpProblem(name=QBCExactModels.MAX_PERIMETER.value, sense=LpMaximize)
    #
    # Variables
    #
    (
        row_choices,
        column_choices,
        cell_choices,
    ) = __row_column_cell_choices_variables(bin_matrix)
    #
    # Objective function
    #
    prob += lpSum(row_choices) + lpSum(column_choices), "max_sum_vertices"
    #
    # Constraints
    #
    __select_cell_select_its_row_column(
        bin_matrix,
        prob,
        row_choices,
        column_choices,
        cell_choices,
    )
    __select_no_more_epsilon_zeros(prob, bin_matrix, epsilon, cell_choices)

    return prob, row_choices, column_choices, cell_choices


# ============================================================================ #
#                                   VARIABLES                                  #
# ============================================================================ #
def __row_column_cell_choices_variables(
    bin_matrix: BinMatrix,
) -> tuple[RowColumnChoicesLP, RowColumnChoicesLP, CellChoicesLP]:
    row_choices = [
        LpVariable(f"xu_{u}", cat=LpInteger, lowBound=0, upBound=1)
        for u in range(bin_matrix.number_of_rows())
    ]
    column_choices = [
        LpVariable(f"xv_{v}", cat=LpInteger, lowBound=0, upBound=1)
        for v in range(bin_matrix.number_of_columns())
    ]
    cell_choices = [
        [
            LpVariable(f"cell_{u}_{v}", cat=LpInteger, lowBound=0, upBound=1)
            for v in range(bin_matrix.number_of_columns())
        ]
        for u in range(bin_matrix.number_of_rows())
    ]

    return row_choices, column_choices, cell_choices


# ============================================================================ #
#                                  CONSTRAINTS                                 #
# ============================================================================ #
def __select_cell_select_its_row_column(
    bin_matrix: BinMatrix,
    prob: LpProblem,
    row_choices: RowColumnChoicesLP,
    column_choices: RowColumnChoicesLP,
    cell_choices: CellChoicesLP,
) -> None:
    # OPTIMIZE forall or sum? Perhaps have a look at robust ILP
    for u in range(bin_matrix.number_of_rows()):
        for v in range(bin_matrix.number_of_columns()):
            prob += (row_choices[u] >= cell_choices[u][v]), f"cell_{u}_{v}_1"

            prob += (column_choices[v] >= cell_choices[u][v]), f"cell_{u}_{v}_2"

            if not bin_matrix[u, v]:
                prob += (
                    (row_choices[u] + column_choices[v] - 1 <= cell_choices[u][v]),
                    f"cell_{u}_{v}_3",
                )


def __select_cell_select_its_row_column_compact(
    bin_matrix: BinMatrix,
    prob: LpProblem,
    row_choices: RowColumnChoicesLP,
    column_choices: RowColumnChoicesLP,
    cell_choices: CellChoicesLP,
) -> None:
    for u in range(bin_matrix.number_of_rows()):
        prob += (
            (
                sum(
                    1 for v in range(bin_matrix.number_of_columns()) if bin_matrix[u, v]
                )
                * row_choices[u]
                >= lpSum(
                    cell_choices[u][v]
                    for v in range(bin_matrix.number_of_columns())
                    if bin_matrix[u, v]
                )
            ),
            f"row_{u}_1",
        )
    for v in range(bin_matrix.number_of_columns()):
        prob += (
            (
                sum(1 for u in range(bin_matrix.number_of_rows()) if bin_matrix[u, v])
                * column_choices[v]
                >= lpSum(
                    cell_choices[u][v]
                    for u in range(bin_matrix.number_of_rows())
                    if bin_matrix[u, v]
                )
            ),
            f"col_{v}_2",
        )
    # OPTIMIZE forall or sum? Perhaps have a look at robust ILP
    for u in range(bin_matrix.number_of_rows()):
        for v in range(bin_matrix.number_of_columns()):
            if not bin_matrix[u, v]:
                prob += (
                    (row_choices[u] + column_choices[v] - 1 <= cell_choices[u][v]),
                    f"cell_{u}_{v}_3",
                )


def __select_no_more_epsilon_zeros(
    prob: LpProblem,
    bin_matrix: BinMatrix,
    epsilon: float,
    cell_choices: CellChoicesLP,
) -> None:
    prob += (
        (
            lpSum(
                (1 - bin_matrix[u, v]) * cell_choices[u][v]
                for u in range(bin_matrix.number_of_rows())
                for v in range(bin_matrix.number_of_columns())
            )
            <= epsilon
            * lpSum(
                cell_choices[u][v]
                for u in range(bin_matrix.number_of_rows())
                for v in range(bin_matrix.number_of_columns())
            )
        ),
        "err_rate",
    )


def __respect_the_minimum_dimensions(
    prob: LpProblem,
    min_number_of_rows: int,
    min_number_of_columns: int,
    row_choices: RowColumnChoicesLP,
    column_choices: RowColumnChoicesLP,
) -> None:
    prob += (
        lpSum(row_choices) >= min_number_of_rows,
        f"The result must have at least {min_number_of_rows} rows",
    )
    prob += (
        lpSum(column_choices) >= min_number_of_columns,
        f"The result must have at least {min_number_of_columns} columns",
    )


def __set_initial_values_from_warm_solution(
    bin_matrix: BinMatrix,
    warm_solution: BinMatrix,
    row_choices: RowColumnChoicesLP,
    column_choices: RowColumnChoicesLP,
    cell_choices: CellChoicesLP,
) -> None:
    row_indices: list[int] = []
    column_indices: list[int] = []
    rows_in_warm_solution = set(warm_solution.row_header())
    for row_idx, row_id in enumerate(bin_matrix.row_header()):
        if row_id in rows_in_warm_solution:
            row_choices[row_idx].setInitialValue(1)
            row_indices.append(row_idx)
        else:
            row_choices[row_idx].setInitialValue(0)

    columns_in_warm_solution = set(warm_solution.column_header())
    for column_idx, column_id in enumerate(bin_matrix.column_header()):
        if column_id in columns_in_warm_solution:
            column_choices[column_idx].setInitialValue(1)
            column_indices.append(column_idx)
        else:
            column_choices[column_idx].setInitialValue(0)

    for i in range(bin_matrix.number_of_rows()):
        for j in range(bin_matrix.number_of_columns()):
            cell_choices[i][j].setInitialValue(0)
    for row_idx, column_idx in itertools.product(row_indices, column_indices):
        cell_choices[row_idx][column_idx].setInitialValue(1)



"""Binary matrix items module."""

from __future__ import annotations

from copy import deepcopy
from typing import TYPE_CHECKING, Hashable

if TYPE_CHECKING:
    from pathlib import Path

from typing import Any, Iterable, Iterator, Literal

import numpy as np
import pandas as pd


# ============================================================================ #
#                                    CLASSES                                   #
# ============================================================================ #
class BinMatrix:
    """Binary matrix.

    The given dataframe must only contain 0-1 values.

    Raises
    ------
    NotABinaryMatrixError
        The non 0-1 values are not replaced and the safe option is set

    """

    @classmethod
    def new_unsafe(
        cls,
        bin_matrix: pd.DataFrame,
        replace_value_map: dict[Any, Literal[0, 1]] | None = None,
    ) -> BinMatrix:
        """Create a new binary matrix without verification.

        Parameters
        ----------
        bin_matrix : pd.DataFrame
            Binary matrix
        replace_value_map : dict
            Replace value map

        Returns
        -------
        BinMatrix
            A binary matrix

        """
        __df: pd.DataFrame = bin_matrix.copy()
        # Replace non 0-1 values
        if replace_value_map is not None:
            if None in replace_value_map:
                replace_value_map[np.nan] = replace_value_map.pop(None)
            __df = __df.replace(replace_value_map)

        self = cls.__new__(cls)

        self.__dataframe = __df
        self.__number_of_ones = int((self.__dataframe == 1).sum().sum())
        return self

    @classmethod
    def from_csv(
        cls,
        csv_path: Path,
        replace_value_map: dict[Any, Literal[0, 1]] | None = None,
    ) -> BinMatrix:
        """Instantiate a binary matrix from CSV file.

        Parameters
        ----------
        csv_path : Path
            CSV path
        replace_value_map : dict[Any, Literal[0, 1]], optional
            Replace value map

        Returns
        -------
        BinMatrix
            A binary matrix

        """
        df: pd.DataFrame = pd.read_csv(csv_path, header=0, index_col=0)
        return cls(df, replace_value_map=replace_value_map)

    @classmethod
    def unsafe_from_csv(
        cls,
        csv_path: Path,
        replace_value_map: dict[Any, Literal[0, 1]] | None = None,
    ) -> BinMatrix:
        """Instantiate a binary matrix from CSV file.

        Parameters
        ----------
        csv_path : Path
            CSV path
        replace_value_map : dict[Any, Literal[0, 1]], optional
            Replace value map

        Returns
        -------
        BinMatrix
            A binary matrix

        """
        df: pd.DataFrame = pd.read_csv(csv_path, header=0, index_col=0)
        return cls.new_unsafe(df, replace_value_map=replace_value_map)

    def __init__(
        self,
        bin_matrix: pd.DataFrame,
        replace_value_map: dict[Any, Literal[0, 1]] | None = None,
    ) -> None:
        """Instantiate a binary matrix.

        It verifies that the given dataframe only contains 0-1 values.
        """
        __df: pd.DataFrame = bin_matrix.copy()
        # Replace non 0-1 values
        if replace_value_map is not None:
            if None in replace_value_map:
                replace_value_map[np.nan] = replace_value_map.pop(None)
            __df = __df.replace(replace_value_map)

        # Verify it takes value in 0, 1
        for value in __df.to_numpy().flatten():
            if value not in (0, 1):
                raise NotABinaryMatrixError(value)

        self.__dataframe = __df
        self.__number_of_ones = int((self.__dataframe == 1).sum().sum())

    def to_dataframe(self) -> pd.DataFrame:
        """Binary matrix dataframe.

        Returns
        -------
        pd.DataFrame
            Binary matrix dataframe

        """
        return self.__dataframe

    def cell_values(self) -> Iterator[tuple[tuple[int, int], Literal[0, 1]]]:
        """Cell values.

        Yields
        ------
        tuple of int
            Cell coordinates
        Literal[0, 1]
            Cell value

        """
        for i in range(self.__dataframe.shape[0]):
            for j in range(self.__dataframe.shape[1]):
                yield (i, j), self.__dataframe.iloc[i, j]

    def dimensions(self) -> tuple[int, int]:
        """Matrix dimensions.

        Returns
        -------
        tuple of int
            Matrix dimensions

        """
        return (
            int(self.__dataframe.shape[0]),
            int(self.__dataframe.shape[1]),
        )

    def number_of_rows(self) -> int:
        """Give the number of rows.

        Returns
        -------
        int
            Number of rows

        """
        return int(self.__dataframe.shape[0])

    def number_of_columns(self) -> int:
        """Give the number of columns.

        Returns
        -------
        int
            Number of columns

        """
        return int(self.__dataframe.shape[1])

    def number_of_cells(self) -> int:
        """Give the number of cells.

        Returns
        -------
        int
            Number of cells

        """
        return int(self.__dataframe.size)

    def row(self, row_id: np.object_) -> pd.Series:
        """Get row.

        Parameters
        ----------
        row_id : np.object_
            Row id

        Returns
        -------
        pd.Series
            A row

        """
        return self.__dataframe.loc[row_id]

    def column(self, column_id: np.object_) -> pd.Series:
        """Get column.

        Parameters
        ----------
        column_id : np.object_
            Column id

        Returns
        -------
        pd.Series
            A column

        """
        return self.__dataframe[column_id]

    def row_degrees(self) -> tuple[int, ...]:
        """Row degrees.

        Returns
        -------
        tuple of int
            Row degrees

        """
        return tuple(self.__dataframe.sum(axis=1))

    def column_degrees(self) -> tuple[int, ...]:
        """Column degrees.

        Returns
        -------
        tuple of int
            Column degrees

        """
        return tuple(self.__dataframe.sum(axis=0))

    def cells_of_interest(
        self,
        value_of_interest: Literal[0, 1],
    ) -> Iterator[tuple[int, int]]:
        """Cell of interest.

        Parameters
        ----------
        value_of_interest : Literal[0, 1]
            Value of interest

        Yields
        ------
        couple of int
            Cell coordinates

        """
        yield from zip(
            *np.nonzero(
                self.__dataframe == value_of_interest,
            ),
        )

    def number_of_zeros(self) -> int:
        """Give the number of zeros.

        Returns
        -------
        int
            Number of zeros

        """
        return self.number_of_cells() - self.__number_of_ones

    def number_of_ones(self) -> int:
        """Give the number of ones.

        Returns
        -------
        int
            Number of ones

        """
        return self.__number_of_ones

    def density(self) -> float:
        """Density.

        Returns
        -------
        float
            Density

        """
        return self.__number_of_ones / self.number_of_cells()

    def sparsity(self) -> float:
        """Sparsity.

        Returns
        -------
        float
            Sparsity

        """
        return 1 - self.density()

    def row_header(self) -> pd.Index:
        """Row header.

        Returns
        -------
        pd.Index
            Row header

        """
        return self.__dataframe.index

    def column_header(self) -> pd.Index:
        """Column header.

        Returns
        -------
        pd.Index
            Column header

        """
        return self.__dataframe.columns

    def iter_rows(self) -> Iterable[tuple[Hashable, pd.Series]]:
        """Iterate over rows.

        Returns
        -------
        Iterable of tuple(hashable, pd.Series)
            A row

        """
        return self.__dataframe.iterrows()

    def iter_columns(self) -> Iterable[tuple[Hashable, pd.Series]]:
        """Iterate over columns.

        Returns
        -------
        Iterable of tuple(hashable, pd.Series)
            A column

        """
        return self.__dataframe.iteritems()

    def __getitem__(self, coordinate: tuple[int, int]) -> Literal[0, 1]:
        """Get cell value.

        Parameters
        ----------
        coordinate : tuple of int
            Cell coordinates

        Returns
        -------
        Literal[0, 1]
            Cell value

        """
        return self.__dataframe.iloc[coordinate[0], coordinate[1]]

    def to_sub_matrix(
        self,
        row_mask: Iterable[bool],
        column_mask: Iterable[bool],
    ) -> BinMatrix:
        """Instantiate a binary matrix from mask.

        Parameters
        ----------
        row_mask : Iterable of bool
            Row mask
        column_mask : Iterable of bool
            Column mask

        Returns
        -------
        BinMatrix
            A binary matrix

        Raises
        ------
        DegeneratedMatrixError
            If the row or the column mask is empty

        """
        row_mask = np.array(row_mask)
        column_mask = np.array(column_mask)
        # pylint: disable=singleton-comparison
        if (row_mask == False).all() or (column_mask == False).all():  # noqa: E712
            raise DegeneratedMatrixError(
                int(sum(row_mask)),
                int(sum(column_mask)),
            )
        return BinMatrix.new_unsafe(self.__dataframe.iloc[row_mask, column_mask])

    def reverse_zeros_and_ones(self) -> BinMatrix:
        """Reverse zeros and ones.

        Returns
        -------
        BinMatrix
            A binary matrix

        """
        return BinMatrix.new_unsafe(
            deepcopy(self.__dataframe).map(lambda x: 1 - x),
        )

    def to_csv(self, csv_path: Path) -> None:
        """Write the binary matrix to a csv file.

        Parameters
        ----------
        csv_path : Path
            Path to the csv file

        Warnings
        --------
        The csv file will be overwritten if it already exists

        """
        self.__dataframe.to_csv(csv_path)


# ============================================================================ #
#                                  EXCEPTIONS                                  #
# ============================================================================ #
class NotABinaryMatrixError(Exception):
    """Not a binary matrix exception."""

    def __init__(self, unexpected_value: Any) -> None:  # noqa: ANN401
        """Initialize the exception."""
        self.__unexpected_value: Any = unexpected_value

    def unexpected_value(self) -> Any:  # noqa: ANN401
        """Unexpected value.

        Returns
        -------
        Any
            Unexpected value

        """
        return self.__unexpected_value

    def __str__(self) -> str:
        """Print the exception message.

        Returns
        -------
        str
            Exception message

        """
        return (
            f"Unexpected value: {self.__unexpected_value}\n"
            "The given CSV file or dataframe must contain"
            " only 0 and 1 values.\n"
            "\tHELP: use the `replace_map_value` argument in the"
            " `BinMatrix.from_csv` class method or in `BinMatrix` constructor."
        )


class DegeneratedMatrixError(Exception):
    """Degenerated matrix exception."""

    def __init__(self, number_of_rows: int, number_of_columns: int) -> None:
        """Initialize the exception."""
        self.__nrow: int = number_of_rows
        self.__ncol: int = number_of_columns

    def number_of_rows(self) -> int:
        """Give the number of rows.

        Returns
        -------
        int
            Number of rows

        """
        return self.__nrow

    def number_of_columns(self) -> int:
        """Give the number of columns.

        Returns
        -------
        int
            Number of columns

        """
        return self.__ncol

    def __str__(self) -> str:
        """Print the exception message.

        Returns
        -------
        str
            Exception message

        """
        return f"The matrix is degenerated ({self.__nrow}x{self.__ncol})."



"""ILP König models."""

from __future__ import annotations

from typing import TYPE_CHECKING

from pulp import LpContinuous, LpMinimize, LpProblem, LpVariable, lpSum

from qbc_dsbm import BCModels

if TYPE_CHECKING:
    from qbc_dsbm.bigraph import ZerosBiGraph

# ============================================================================ #
#                                     TYPES                                    #
# ============================================================================ #
VertexChoicesLP = list[LpVariable]


# ============================================================================ #
#                      LP MODEL - KÖNIG THEOREM CLASSICAL                      #
# ============================================================================ #
def konig_v(
    bigraph: ZerosBiGraph,
) -> tuple[LpProblem, VertexChoicesLP, VertexChoicesLP]:
    """König_V model.

    Parameters
    ----------
    bigraph : BiGraph
        Bigraph

    Returns
    -------
    LpProblem
        PuLP problem
    VertexChoicesLP
        U vertex choices
    VertexChoicesLP
        V vertex choices

    """
    prob = LpProblem(name=BCModels.KONIG_V.value, sense=LpMinimize)
    #
    # Variables
    #
    u_vertex_choices, v_vertex_choices = __vertex_choices_variables(bigraph)
    #
    # Objective function
    #
    prob += lpSum(u_vertex_choices) + lpSum(v_vertex_choices), "min_vertices"
    #
    # Constraints
    #
    __delete_row_or_col_contains_zero(prob, bigraph, u_vertex_choices, v_vertex_choices)

    return prob, u_vertex_choices, v_vertex_choices


# ============================================================================ #
#                         LP MODEL - KÖNIG WITH DEGREES                        #
# ============================================================================ #
def konig_e(
    bigraph: ZerosBiGraph,
) -> tuple[LpProblem, VertexChoicesLP, VertexChoicesLP]:
    """König_E model.

    Parameters
    ----------
    bigraph : BiGraph
        Bigraph

    Returns
    -------
    LpProblem
        PuLP problem
    VertexChoicesLP
        U vertex choices
    VertexChoicesLP
        V vertex choices

    """
    prob = LpProblem(name=BCModels.KONIG_E.value, sense=LpMinimize)
    #
    # Variables
    #
    u_vertex_choices, v_vertex_choices = __vertex_choices_variables(bigraph)
    #
    # Objective function
    #
    prob += (
        (
            lpSum(
                (
                    var * deg
                    for var, deg in zip(
                        u_vertex_choices,
                        bigraph.u_bipart_anti_degrees(),
                    )
                ),
            )
            + lpSum(
                (
                    var * deg
                    for var, deg in zip(
                        v_vertex_choices,
                        bigraph.v_bipart_anti_degrees(),
                    )
                ),
            )
        ),
        "min_lost_of_ones",
    )
    #
    # Constraints
    #
    __delete_row_or_col_contains_zero(prob, bigraph, u_vertex_choices, v_vertex_choices)

    return prob, u_vertex_choices, v_vertex_choices


# ============================================================================ #
#                                   VARIABLES                                  #
# ============================================================================ #
def __vertex_choices_variables(
    bigraph: ZerosBiGraph,
) -> tuple[VertexChoicesLP, VertexChoicesLP]:
    return (
        [
            LpVariable(f"xu_{u}", cat=LpContinuous, lowBound=0, upBound=1)
            for u in range(bigraph.card_u_bipart())
        ],
        [
            LpVariable(f"xv_{v}", cat=LpContinuous, lowBound=0, upBound=1)
            for v in range(bigraph.card_v_bipart())
        ],
    )


# ============================================================================ #
#                                  CONSTRAINTS                                 #
# ============================================================================ #
def __delete_row_or_col_contains_zero(
    prob: LpProblem,
    bigraph: ZerosBiGraph,
    u_vertex_choices: VertexChoicesLP,
    v_vertex_choices: VertexChoicesLP,
) -> None:
    for u, v in bigraph.edges():
        prob += (u_vertex_choices[u] + v_vertex_choices[v] >= 1), f"edge_{u}_{v}"



"""Bigraph module."""

from __future__ import annotations

from typing import TYPE_CHECKING, Iterable, cast

import numpy as np
from nptyping import NDArray, Shape, UnsignedInteger

if TYPE_CHECKING:
    from qbc_dsbm.binary_matrix import BinMatrix


# ============================================================================ #
#                                    CLASSES                                   #
# ============================================================================ #
class ZerosBiGraph:
    """Bigraph class.

    The edges represent zeros in the binary matrix.
    """

    def __init__(
        self,
        edges: Iterable[tuple[int, int]],
        u_bipart_size: int,
        v_bipart_size: int,
    ) -> None:
        """Initialize the bigraph."""
        self.__edges: tuple[tuple[int, int], ...] = tuple(edges)

        self.__u_bipart_degrees = np.zeros((u_bipart_size,), dtype=UnsignedInteger)
        self.__v_bipart_degrees = np.zeros((v_bipart_size,), dtype=UnsignedInteger)

        for u, v in self.__edges:
            self.__u_bipart_degrees[u] += 1
            self.__v_bipart_degrees[v] += 1

    @classmethod
    def from_bin_matrix(cls, bin_matrix: BinMatrix) -> ZerosBiGraph:
        """Instantiate a bigraph from a binary matrix.

        Parameters
        ----------
        bin_matrix : BinMatrix
            Binary matrix

        Returns
        -------
        BiGraph
            A bigraph

        """
        return cls(
            bin_matrix.cells_of_interest(0),
            bin_matrix.number_of_rows(),
            bin_matrix.number_of_columns(),
        )

    def card_u_bipart(self) -> int:
        """Cardinality of the first bipartite.

        Returns
        -------
        int
            Cardinality of bipart U

        """
        return len(self.__u_bipart_degrees)

    def card_v_bipart(self) -> int:
        """Cardinality of the second bipartite.

        Returns
        -------
        int
            Cardinality of bipart V

        """
        return len(self.__v_bipart_degrees)

    def u_bipart_degrees(self) -> NDArray[Shape[1], UnsignedInteger]:
        """First bipartite degrees.

        Returns
        -------
        vector of int
            First bipartite degrees

        """
        return self.__u_bipart_degrees

    def v_bipart_degrees(self) -> NDArray[Shape[1], UnsignedInteger]:
        """Second bipartite degrees.

        Returns
        -------
        vector of int
            Second bipartite degrees

        """
        return self.__v_bipart_degrees

    def u_bipart_anti_degrees(self) -> NDArray[Shape[1], UnsignedInteger]:
        """First bipartite anti-degrees.

        Anti-degrees are the number of missing edges on the bipartite
        to make the bigraph complete.

        Returns
        -------
        vector of int
            First bipartite anti-degrees

        """
        return cast(
            NDArray[Shape["1"], UnsignedInteger],
            len(self.__v_bipart_degrees) - self.__u_bipart_degrees,
        )

    def v_bipart_anti_degrees(self) -> NDArray[Shape[1], UnsignedInteger]:
        """Second bipartite anti-degrees.

        Anti-degrees are the number of missing edges on the bipartite
        to make the bigraph complete.

        Returns
        -------
        vector of int
            Second bipartite anti-degrees

        """
        return cast(
            NDArray[Shape["1"], UnsignedInteger],
            len(self.__u_bipart_degrees) - self.__v_bipart_degrees,
        )

    def edges(self) -> tuple[tuple[int, int], ...]:
        """Edges.

        Returns
        -------
        tuple of tuple of int
            Edges

        """
        return self.__edges



"""Loggers module."""

from __future__ import annotations

import logging
import sys
from pathlib import Path

import yaml  # type: ignore[import-untyped]

from qbc_dsbm.binary_matrix import BinMatrixStats
from qbc_dsbm.io_config import IOConfig, IOConfigDivCqr
from qbc_dsbm.models.views import (
    ModelLPMILPStats,
    ModelLPStats,
    ModelSolveStats,
    ModelType,
)

_LOGGER = logging.getLogger(__name__)


# ============================================================================ #
#                                    CLASSES                                   #
# ============================================================================ #
class UniqueStrategyLogger:
    """Logger for the unique strategy."""

    __IO_CONFIG_KEY = "io_config"
    """IO config key."""

    __INIT_BIN_MATRIX_KEY = "init_bin_matrix"
    """Initial binary matrix key."""

    __MODEL_STATS_KEY = "model_solve_stats"
    """Model solve statistics key."""

    __RESULT_BIN_MATRIX_KEY = "result_bin_matrix"
    """Result binary matrix key."""

    __YAML_FILE: Path = Path("unique_strategy.yaml")
    """Unique strategy YAML file."""

    @classmethod
    def yaml_file(cls) -> Path:
        """Return the YAML file.

        Returns
        -------
        Path
            YAML file

        """
        return cls.__YAML_FILE

    @classmethod
    def from_yaml(cls, directory: Path) -> UniqueStrategyLogger:
        """Load the logger from the YAML file.

        Parameters
        ----------
        directory : Path
            Directory

        Returns
        -------
        UniqueStrategyLogger
            Unique strategy logger

        """
        with (directory / cls.yaml_file()).open(encoding="utf-8") as stream:
            yaml_dict = yaml.safe_load(stream)
        model_solve_stats: ModelSolveStats
        match ModelSolveStats.model_type_from_dict(yaml_dict[cls.__MODEL_STATS_KEY]):
            case ModelType.MILP:
                model_solve_stats = ModelLPMILPStats.from_dict(
                    yaml_dict[cls.__MODEL_STATS_KEY],
                )
            case ModelType.LP:
                model_solve_stats = ModelLPStats.from_dict(
                    yaml_dict[cls.__MODEL_STATS_KEY],
                )
            case _:
                _LOGGER.critical("Cannot recognize model type")
                sys.exit(1)
        return cls(
            io_config=IOConfig.from_dict(yaml_dict[cls.__IO_CONFIG_KEY]),
            init_bin_matrix=BinMatrixStats.from_dict(
                yaml_dict[cls.__INIT_BIN_MATRIX_KEY],
            ),
            model_solve_stats=model_solve_stats,
            result_bin_matrix=BinMatrixStats.from_dict(
                yaml_dict[cls.__RESULT_BIN_MATRIX_KEY],
            ),
        )

    def __init__(
        self,
        io_config: IOConfig,
        init_bin_matrix: BinMatrixStats,
        model_solve_stats: ModelSolveStats | None = None,
        result_bin_matrix: BinMatrixStats | None = None,
    ) -> None:
        """Initialize the logger."""
        self.__io_config: IOConfig = io_config
        self.__init_bin_matrix: BinMatrixStats = init_bin_matrix
        self.__model_solve_stats: ModelSolveStats | None = model_solve_stats
        self.__result_bin_matrix: BinMatrixStats | None = result_bin_matrix

    def io_config(self) -> IOConfig:
        """IO config.

        Returns
        -------
        IOConfig
            IO config

        """
        return self.__io_config

    def init_bin_matrix(self) -> BinMatrixStats:
        """Give the initial binary matrix.

        Returns
        -------
        BinMatrixStats
            Initial binary matrix

        """
        return self.__init_bin_matrix

    def model_solve_stats(self) -> ModelSolveStats | None:
        """Give the solve statistics.

        Returns
        -------
        ModelSolveStats | None
            Solve statistics

        """
        return self.__model_solve_stats

    def result_bin_matrix(self) -> BinMatrixStats | None:
        """Give the resulting binary matrix.

        Returns
        -------
        BinMatrixStats | None
            Result binary matrix

        """
        return self.__result_bin_matrix

    def add_solve_stats(self, model_solve_stats: ModelSolveStats) -> None:
        """Add Solve statistics.

        Parameters
        ----------
        model_solve_stats : ModelSolveStats
            Solve statistics

        """
        self.__model_solve_stats = model_solve_stats

    def add_result_bin_matrix(self, result_bin_matrix: BinMatrixStats) -> None:
        """Add result binary matrix.

        Parameters
        ----------
        result_bin_matrix : BinMatrixStats
            Result binary matrix

        """
        self.__result_bin_matrix = result_bin_matrix

    def write_yaml(self, directory: Path) -> None:
        """Write YAML.

        Parameters
        ----------
        directory : Path
            Output directory

        """
        with (directory / self.yaml_file()).open("w", encoding="utf-8") as yaml_out:
            yaml_out.write(
                yaml.dump(
                    {
                        self.__IO_CONFIG_KEY: (self.__io_config.to_dict()),
                        self.__INIT_BIN_MATRIX_KEY: (self.__init_bin_matrix.to_dict()),
                        self.__MODEL_STATS_KEY: (
                            self.__model_solve_stats.to_dict()
                            if self.__model_solve_stats is not None
                            else None
                        ),
                        self.__RESULT_BIN_MATRIX_KEY: (
                            self.__result_bin_matrix.to_dict()
                            if self.__result_bin_matrix is not None
                            else None
                        ),
                    },
                    sort_keys=False,
                ),
            )


class DivCqrStrategyLogger:
    """Logger for the divide-and-conquer strategy."""

    __IO_CONFIG_KEY = "io_config"
    """IO config key."""

    __INIT_BIN_MATRIX_KEY = "init_bin_matrix"
    """Initial binary matrix key."""

    __HEURISTIC_SOLVE_STATS_KEY = "heuristic_solve_stats"
    """Heuristic Solve statistics key."""

    __HEURISTIC_RESULT_BIN_MATRIX_KEY = "heuristic_result_bin_matrix"
    """Heuristic result binary matrix key."""

    __EXACT_SOLVE_STATS_KEY = "exact_solve_stats"
    """Exact Solve statistics key."""

    __EXACT_RESULT_BIN_MATRIX_KEY = "exact_result_bin_matrix"
    """Exact result binary matrix key."""

    __YAML_FILE: Path = Path("divcqr_strategy.yaml")
    """DivCqr strategy YAML file."""

    @classmethod
    def yaml_file(cls) -> Path:
        """Return the YAML file.

        Returns
        -------
        Path
            YAML file

        """
        return cls.__YAML_FILE

    @classmethod
    def from_yaml(cls, directory: Path) -> DivCqrStrategyLogger:
        """Instantiate from YAML.

        Parameters
        ----------
        directory : Path
            Directory

        Returns
        -------
        DivCqrStrategyLogger
            DivCqr strategy logger

        """
        with (directory / cls.yaml_file()).open(encoding="utf-8") as stream:
            yaml_dict = yaml.safe_load(stream)
        heuristic_solve_stats: ModelSolveStats
        match ModelSolveStats.model_type_from_dict(
            yaml_dict[cls.__HEURISTIC_SOLVE_STATS_KEY],
        ):
            case ModelType.MILP:
                heuristic_solve_stats = ModelLPMILPStats.from_dict(
                    yaml_dict[cls.__HEURISTIC_SOLVE_STATS_KEY],
                )
            case ModelType.LP:
                heuristic_solve_stats = ModelLPStats.from_dict(
                    yaml_dict[cls.__HEURISTIC_SOLVE_STATS_KEY],
                )
            case _:
                _LOGGER.critical("Cannot recognize exact model type")
                sys.exit(1)
        return cls(
            io_config=IOConfigDivCqr.from_dict(yaml_dict[cls.__IO_CONFIG_KEY]),
            init_bin_matrix=BinMatrixStats.from_dict(
                yaml_dict[cls.__INIT_BIN_MATRIX_KEY],
            ),
            heuristic_solve_stats=heuristic_solve_stats,
            heuristic_result_bin_matrix=BinMatrixStats.from_dict(
                yaml_dict[cls.__HEURISTIC_RESULT_BIN_MATRIX_KEY],
            ),
            exact_solve_stats=ModelLPMILPStats.from_dict(
                yaml_dict[cls.__EXACT_SOLVE_STATS_KEY],
            ),
            exact_result_bin_matrix=BinMatrixStats.from_dict(
                yaml_dict[cls.__EXACT_RESULT_BIN_MATRIX_KEY],
            ),
        )

    def __init__(  # noqa: PLR0913
        self,
        io_config: IOConfigDivCqr,
        init_bin_matrix: BinMatrixStats,
        heuristic_solve_stats: ModelSolveStats | None = None,
        heuristic_result_bin_matrix: BinMatrixStats | None = None,
        exact_solve_stats: ModelLPMILPStats | None = None,
        exact_result_bin_matrix: BinMatrixStats | None = None,
    ) -> None:
        """Initialize the logger."""
        self.__io_config: IOConfigDivCqr = io_config
        self.__init_bin_matrix: BinMatrixStats = init_bin_matrix
        self.__heuristic_solve_stats: ModelSolveStats | None = heuristic_solve_stats
        self.__heuristic_result_bin_matrix: BinMatrixStats | None = (
            heuristic_result_bin_matrix
        )
        self.__exact_solve_stats: ModelLPMILPStats | None = exact_solve_stats
        self.__exact_result_bin_matrix: BinMatrixStats | None = exact_result_bin_matrix

    def io_config(self) -> IOConfigDivCqr:
        """IO config.

        Returns
        -------
        IOConfigDivCqr
            IO config

        """
        return self.__io_config

    def init_bin_matrix(self) -> BinMatrixStats:
        """Give the initial binary matrix.

        Returns
        -------
        BinMatrixStats
            Initial binary matrix

        """
        return self.__init_bin_matrix

    def heuristic_solve_stats(self) -> ModelSolveStats | None:
        """Give the heuristic Solve statistics.

        Returns
        -------
        ModelSolveStats | None
            Heuristic Solve statistics

        """
        return self.__heuristic_solve_stats

    def heuristic_result_bin_matrix(self) -> BinMatrixStats | None:
        """Give the heuristic resulting binary matrix.

        Returns
        -------
        BinMatrixStats | None
            Heuristic result binary matrix

        """
        return self.__heuristic_result_bin_matrix

    def exact_solve_stats(self) -> ModelLPMILPStats | None:
        """Give the exact solve statistics.

        Returns
        -------
        ModelLPMILPStats | None
            Exact solve statistics

        """
        return self.__exact_solve_stats

    def exact_result_bin_matrix(self) -> BinMatrixStats | None:
        """Give the exact resulting binary matrix.

        Returns
        -------
        BinMatrixStats | None
            Exact result binary matrix

        """
        return self.__exact_result_bin_matrix

    def add_heuristic_solve_stats(
        self,
        heuristic_solve_stats: ModelSolveStats,
    ) -> None:
        """Add heuristic solve statistics.

        Parameters
        ----------
        heuristic_solve_stats : ModelSolveStats
            Heuristic solve statistics

        """
        self.__heuristic_solve_stats = heuristic_solve_stats

    def add_heuristic_result_bin_matrix(
        self,
        heuristic_result_bin_matrix: BinMatrixStats,
    ) -> None:
        """Add heuristic result binary matrix.

        Parameters
        ----------
        heuristic_result_bin_matrix : BinMatrixStats
            Heuristic result binary matrix

        """
        self.__heuristic_result_bin_matrix = heuristic_result_bin_matrix

    def add_exact_solve_stats(self, exact_solve_stats: ModelLPMILPStats) -> None:
        """Add exact solve statistics.

        Parameters
        ----------
        exact_solve_stats : ModelLPMILPStats
            Exact solve statistics

        """
        self.__exact_solve_stats = exact_solve_stats

    def add_exact_result_bin_matrix(
        self,
        exact_result_bin_matrix: BinMatrixStats,
    ) -> None:
        """Add exact result binary matrix.

        Parameters
        ----------
        exact_result_bin_matrix : BinMatrixStats
            Exact result binary matrix

        """
        self.__exact_result_bin_matrix = exact_result_bin_matrix

    def write_yaml(self, directory: Path) -> None:
        """Write YAML.

        Parameters
        ----------
        directory : Path
            Output directory

        """
        with (directory / self.yaml_file()).open("w", encoding="utf-8") as yaml_out:
            yaml_out.write(
                yaml.dump(
                    {
                        self.__IO_CONFIG_KEY: (self.__io_config.to_dict()),
                        self.__INIT_BIN_MATRIX_KEY: (self.__init_bin_matrix.to_dict()),
                        self.__HEURISTIC_SOLVE_STATS_KEY: (
                            self.__heuristic_solve_stats.to_dict()
                            if self.__heuristic_solve_stats is not None
                            else None
                        ),
                        self.__HEURISTIC_RESULT_BIN_MATRIX_KEY: (
                            self.__heuristic_result_bin_matrix.to_dict()
                            if self.__heuristic_result_bin_matrix is not None
                            else None
                        ),
                        self.__EXACT_SOLVE_STATS_KEY: (
                            self.__exact_solve_stats.to_dict()
                            if self.__exact_solve_stats is not None
                            else None
                        ),
                        self.__EXACT_RESULT_BIN_MATRIX_KEY: (
                            self.__exact_result_bin_matrix.to_dict()
                            if self.__exact_result_bin_matrix is not None
                            else None
                        ),
                    },
                    sort_keys=False,
                ),
            )



"""Problem solvers."""

from __future__ import annotations

import logging
from typing import TYPE_CHECKING, Iterable, Literal

if TYPE_CHECKING:
    from pulp import LpVariable

    from qbc_dsbm.models.views import ModelLPMILPStats, ModelSolveStats

from qbc_dsbm import BCModels, QBCExactModels, QBCHeuristicModels
from qbc_dsbm.bigraph import ZerosBiGraph
from qbc_dsbm.binary_matrix import BinMatrix, DegeneratedMatrixError
from qbc_dsbm.models.exact_dsbm import (
    max_ones,
    max_ones_compact,
    max_perimeter,
    max_surface,
)
from qbc_dsbm.models.heuristic_dsbm import (
    kp_qb,
    min_del_cols,
    min_del_cols_relax,
    min_del_ones,
    min_del_rc,
    min_del_rows,
    min_del_rows_relax,
)
from qbc_dsbm.models.konig import konig_e, konig_v
from qbc_dsbm.models.solvers import NoSolutionError, Solver, SolverTimeLimitReachedError

LOGGER = logging.getLogger(__name__)


# ============================================================================ #
#                                   FUNCTIONS                                  #
# ============================================================================ #
# ---------------------------------------------------------------------------- #
#                                  Exact DSBM                                  #
# ---------------------------------------------------------------------------- #
# REFACTOR use exact config
# REFACTOR each model should have its own command / config
def solve_exact_dsbm(
    bin_matrix: BinMatrix,
    model: QBCExactModels,
    epsilon: float,
    solver: Solver,
    min_number_of_rows: int = 0,
    min_number_of_columns: int = 0,  # XXX optional constraints
    warm_solution: BinMatrix | None = None,
    check_satisfiability: bool = False,  # noqa: FBT001, FBT002
) -> tuple[BinMatrix, ModelLPMILPStats]:
    """Solve exact DSBM problem.

    Parameters
    ----------
    bin_matrix : BinMatrix
        Binary matrix
    model : QBCExactModels
        Exact model
    epsilon : float
        Error rate
    solver : Solver
        Solver to use
    min_number_of_rows : int
        Minimum number of rows
    min_number_of_columns : int
        Minimum number of columns
    warm_solution : BinMatrix
        Warm solution
    check_satisfiability : bool
        Check satisfiability, do not solve an optimization problem

    Returns
    -------
    BinMatrix
        Sub-binary matrix
    SolveStats
        Solve statistics

    Raises
    ------
    SolveError
        The problem could not be solved
    ModelFindsDegeneratedMatrixError
        The model finds degenerated matrix

    """
    match model:
        case QBCExactModels.MAX_ONES:
            prob, row_choices, col_choices, _ = max_ones(bin_matrix, epsilon)
            vars_are_integers = True
        case QBCExactModels.MAX_ONES_COMPACT:
            prob, row_choices, col_choices, _ = max_ones_compact(
                bin_matrix,
                epsilon,
                min_number_of_rows=min_number_of_rows,
                min_number_of_columns=min_number_of_columns,
                warm_solution=warm_solution,
                check_satisfiability=check_satisfiability,
            )
            vars_are_integers = True
        case QBCExactModels.MAX_PERIMETER:
            prob, row_choices, col_choices, _ = max_perimeter(bin_matrix, epsilon)
            vars_are_integers = True
        case QBCExactModels.MAX_SURFACE:
            prob, row_choices, col_choices, _ = max_surface(bin_matrix, epsilon)
            vars_are_integers = True

    try:
        solve_stats = solver.solve_milp(prob, warm_start=(warm_solution is not None))
    except NoSolutionError as no_sol_err:
        raise SolveError(no_sol_err) from no_sol_err
    except SolverTimeLimitReachedError as time_lim_err:
        raise SolveError(time_lim_err) from time_lim_err

    row_mask, col_mask = __row_col_vertex_choices_to_masks(
        bin_matrix.number_of_rows(),
        bin_matrix.number_of_columns(),
        vars_are_integers,
        row_choices,
        col_choices,
        (False, True),
    )

    try:
        return bin_matrix.to_sub_matrix(row_mask, col_mask), solve_stats
    except DegeneratedMatrixError as degen_error:
        raise ModelFindsDegeneratedMatrixError(
            solve_stats,
            degen_error,
        ) from degen_error


# ---------------------------------------------------------------------------- #
#                                     König                                    #
# ---------------------------------------------------------------------------- #
def solve_konig(
    bin_matrix: BinMatrix,
    model: BCModels,
    solver: Solver,
) -> tuple[BinMatrix, ModelLPMILPStats]:
    """Solve bi-clique König problems.

    Parameters
    ----------
    bin_matrix : BinMatrix
        Binary matrix
    model : BCModels
        Biclique model
    solver : Solver
        Solver to use

    Returns
    -------
    BinMatrix
        Sub-binary matrix
    SolveStats
        Solve statistics

    Raises
    ------
    SolveError
        The problem could not be solved
    ModelFindsDegeneratedMatrixError
        The model finds degenerated matrix

    """
    bigraph = ZerosBiGraph.from_bin_matrix(bin_matrix)

    match model:
        case BCModels.KONIG_V:
            prob, u_vertex_choices, v_vertex_choices = konig_v(bigraph)
            vars_are_integers = True
        case BCModels.KONIG_E:
            prob, u_vertex_choices, v_vertex_choices = konig_e(bigraph)
            vars_are_integers = True

    try:
        solve_stats = solver.solve_milp(prob)
    except NoSolutionError as no_sol_err:
        raise SolveError(no_sol_err) from no_sol_err
    except SolverTimeLimitReachedError as time_lim_err:
        raise SolveError(time_lim_err) from time_lim_err

    row_mask, col_mask = __row_col_vertex_choices_to_masks(
        bigraph.card_u_bipart(),
        bigraph.card_v_bipart(),
        vars_are_integers,
        u_vertex_choices,
        v_vertex_choices,
        (True, False),
    )

    try:
        return bin_matrix.to_sub_matrix(row_mask, col_mask), solve_stats
    except DegeneratedMatrixError as degen_error:
        raise ModelFindsDegeneratedMatrixError(
            solve_stats,
            degen_error,
        ) from degen_error


# ---------------------------------------------------------------------------- #
#                                   Heuristic                                  #
# ---------------------------------------------------------------------------- #
def solve_heuristic_dsbm(  # noqa: C901
    bin_matrix: BinMatrix,
    model: QBCHeuristicModels,
    epsilon: float,
    solver: Solver,
) -> tuple[BinMatrix, ModelSolveStats]:
    """Solve heuristic DSBM problem.

    Parameters
    ----------
    bin_matrix : BinMatrix
        Binary matrix
    model : QBCHeuristicModels
        Quasi bi-clique heuristic model
    epsilon : float
        Error rate
    solver : Solver
        Solver to use

    Returns
    -------
    BinMatrix
        Sub-binary matrix
    ModelSolveStats
        Solve statistics

    Raises
    ------
    SolveError
        The problem could not be solved
    ModelFindsDegeneratedMatrixError
        The model finds degenerated matrix

    """
    bigraph = ZerosBiGraph.from_bin_matrix(bin_matrix)

    is_milp = True
    u_vertex_choices: Iterable[LpVariable] | None
    v_vertex_choices: Iterable[LpVariable] | None
    match model:
        case QBCHeuristicModels.MIN_DEL_ONES:
            prob, u_vertex_choices, v_vertex_choices, _ = min_del_ones(bigraph, epsilon)
            vars_are_integers = True
        case QBCHeuristicModels.MIN_DEL_RC:
            prob, u_vertex_choices, v_vertex_choices, _ = min_del_rc(bigraph, epsilon)
            vars_are_integers = True
        case QBCHeuristicModels.MIN_DEL_ROWS:
            prob, u_vertex_choices, _ = min_del_rows(bigraph, epsilon)
            v_vertex_choices = None
            vars_are_integers = True
        case QBCHeuristicModels.MIN_DEL_COLS:
            prob, v_vertex_choices, _ = min_del_cols(bigraph, epsilon)
            u_vertex_choices = None
            vars_are_integers = True
        case QBCHeuristicModels.MIN_DEL_ROWS_RELAX:
            is_milp = False
            prob, u_vertex_choices, _ = min_del_rows_relax(bigraph, epsilon)
            v_vertex_choices = None
            vars_are_integers = False
        case QBCHeuristicModels.MIN_DEL_COLS_RELAX:
            is_milp = False
            prob, v_vertex_choices, _ = min_del_cols_relax(bigraph, epsilon)
            u_vertex_choices = None
            vars_are_integers = False
        case QBCHeuristicModels.KP_QB:
            prob, u_vertex_choices, v_vertex_choices = kp_qb(bigraph, epsilon)
            vars_are_integers = True

    try:
        solve_stats = solver.solve_milp(prob) if is_milp else solver.solve_lp(prob)
    except NoSolutionError as no_sol_err:
        raise SolveError(no_sol_err) from no_sol_err
    except SolverTimeLimitReachedError as time_lim_err:
        raise SolveError(time_lim_err) from time_lim_err

    row_mask, col_mask = __row_col_vertex_choices_to_masks(
        bigraph.card_u_bipart(),
        bigraph.card_v_bipart(),
        vars_are_integers,
        u_vertex_choices,
        v_vertex_choices,
        (True, False),
    )

    try:
        return bin_matrix.to_sub_matrix(row_mask, col_mask), solve_stats
    except DegeneratedMatrixError as degen_error:
        raise ModelFindsDegeneratedMatrixError(
            solve_stats,
            degen_error,
        ) from degen_error


# ---------------------------------------------------------------------------- #
#                             LP Variables To Masks                            #
# ---------------------------------------------------------------------------- #
def __row_col_vertex_choices_to_masks(  # noqa: PLR0913
    n_row: int,
    n_col: int,
    vars_are_integers: bool,  # noqa: FBT001
    row_choices: Iterable[LpVariable] | None,
    col_choices: Iterable[LpVariable] | None,
    keep: tuple[bool, bool],
) -> tuple[list[bool], list[bool]]:
    masks: tuple[list[bool], list[bool]] = ([], [])
    for k, (choices, dim) in enumerate(
        (
            (row_choices, n_row),
            (col_choices, n_col),
        ),
    ):
        if choices is None:
            for _ in range(dim):
                masks[k].append(keep[0])
        else:
            for var in choices:
                masks[k].append(
                    keep[
                        __lp_var_value_to_bin_value(var)
                        if vars_are_integers
                        else __continuous_var_value_to_bin_value(var)
                    ],
                )
    return masks


def __lp_var_value_to_bin_value(var: LpVariable) -> Literal[0, 1]:
    if var.value() is None:
        LOGGER.warning("None variable value replaced by 1")
        var.setInitialValue(1)
    if var.value() == 1:
        return 1
    if var.value() == 0:
        return 0
    msg = f"Invalid bool value {var.value()}"
    raise ValueError(msg)


def __continuous_var_value_to_bin_value(var: LpVariable) -> Literal[0, 1]:
    if var.value() is None:
        LOGGER.warning("None variable value replaced by 1")
        var.setInitialValue(1)
    if var.value() == 1:
        return 1
    return 0


# ============================================================================ #
#                                   EXCEPTION                                  #
# ============================================================================ #
class SolveError(Exception):
    """Solve exception."""

    def __init__(self, error: SolverTimeLimitReachedError | NoSolutionError) -> None:
        """Initialize the exception."""
        self.__error: SolverTimeLimitReachedError | NoSolutionError = error

    def error(self) -> SolverTimeLimitReachedError | NoSolutionError:
        """Error.

        Returns
        -------
        SolverTimeLimitReachedError | NoSolutionError
            Error

        """
        return self.__error

    def __str__(self) -> str:
        """Print the exception message.

        Returns
        -------
        str
            Exception message

        """
        return f"Problem solving failed:\n{self.__error}\n"


class ModelFindsDegeneratedMatrixError(Exception):
    """Model finds degenerated matrix exception."""

    def __init__(
        self,
        solve_stats: ModelSolveStats,
        degen_error: DegeneratedMatrixError,
    ) -> None:
        """Initialize the exception."""
        self.__solve_stats: ModelSolveStats = solve_stats
        self.__degen_error: DegeneratedMatrixError = degen_error

    def solve_stats(self) -> ModelSolveStats:
        """Solve statistics.

        Returns
        -------
        SolveStats
            Solve statistics

        """
        return self.__solve_stats

    def degen_error(self) -> DegeneratedMatrixError:
        """Degenerated matrix error.

        Returns
        -------
        DegeneratedMatrixError
            Degenerated matrix error

        """
        return self.__degen_error

    def __str__(self) -> str:
        """Print the exception message.

        Returns
        -------
        str
            Exception message

        """
        return (
            f"The model finds degenerated matrix:\n"
            f"{self.__degen_error}\n"
            f"See the solve statistics:\n"
            f"{self.__solve_stats}"
        )



"""Solving strategies."""

from __future__ import annotations

from typing import TYPE_CHECKING, cast

from qbc_dsbm import BCModels, QBCExactModels, QBCHeuristicModels
from qbc_dsbm.binary_matrix import BinMatrix, BinMatrixStats

if TYPE_CHECKING:
    from qbc_dsbm.io_config import IOConfig, IOConfigDivCqr
    from qbc_dsbm.models.views import ModelSolveStats

import logging

from qbc_dsbm.loggers import DivCqrStrategyLogger, UniqueStrategyLogger
from qbc_dsbm.models.views import ModelLPMILPStats
from qbc_dsbm.solve_problems import (
    ModelFindsDegeneratedMatrixError,
    SolveError,
    solve_exact_dsbm,
    solve_heuristic_dsbm,
    solve_konig,
)

_LOGGER = logging.getLogger(__name__)


# ============================================================================ #
#                                   FUNCTIONS                                  #
# ============================================================================ #
def unique_strategy(io_config: IOConfig) -> tuple[BinMatrix, UniqueStrategyLogger]:
    """Solve the unique strategy.

    Parameters
    ----------
    io_config : IOConfig
        IO config

    Returns
    -------
    BinMatrix
        Sub-binary matrix
    UniqueStrategyLogger
        Logger

    Raises
    ------
    UniqueStrategyError
        The unique strategy failed

    """
    init_bin_matrix = BinMatrix.from_csv(io_config.csv_path())
    logger = UniqueStrategyLogger(
        io_config,
        BinMatrixStats.from_bin_matrix(init_bin_matrix),
    )

    _LOGGER.info("Input / Output config:\n%s\n", io_config)
    _LOGGER.info(
        "Initial binary matrix stats:\n%s\n",
        str(logger.init_bin_matrix()),
    )

    model = io_config.model()
    solve_stats: ModelSolveStats
    try:
        if model in QBCExactModels:
            result_bin_matrix, solve_stats = solve_exact_dsbm(
                init_bin_matrix,
                cast(QBCExactModels, model),
                io_config.epsilon(),
                io_config.solver(),
                min_number_of_rows=io_config.min_number_of_rows(),
                min_number_of_columns=io_config.min_number_of_columns(),
                check_satisfiability=io_config.check_satisfiability(),
            )
        elif model in BCModels:
            result_bin_matrix, solve_stats = solve_konig(
                init_bin_matrix,
                cast(BCModels, model),
                io_config.solver(),
            )
        elif model in QBCHeuristicModels:
            result_bin_matrix, solve_stats = solve_heuristic_dsbm(
                init_bin_matrix,
                cast(QBCHeuristicModels, model),
                io_config.epsilon(),
                io_config.solver(),
            )
    except (
        ModelFindsDegeneratedMatrixError,
        SolveError,
    ) as error:
        if isinstance(error, ModelFindsDegeneratedMatrixError):
            logger.add_solve_stats(error.solve_stats())
        elif isinstance(error, SolveError):
            logger.add_solve_stats(error.error().solve_stats())
        _LOGGER.info(
            "Model solve stats:\n%s\n",
            logger.model_solve_stats(),
        )
        _LOGGER.critical(str(error))
        raise UniqueStrategyError(logger) from error
    logger.add_solve_stats(solve_stats)
    logger.add_result_bin_matrix(
        BinMatrixStats.from_bin_matrix(result_bin_matrix),
    )
    _LOGGER.info(
        "Model solve stats:\n%s\n",
        logger.model_solve_stats(),
    )
    _LOGGER.info(
        "Result binary matrix stats:\n%s\n",
        str(logger.result_bin_matrix()),
    )
    return result_bin_matrix, logger


def divide_and_conquer_strategy(
    io_config: IOConfigDivCqr,
) -> tuple[BinMatrix, DivCqrStrategyLogger]:
    """Solve the divide-and-conquer strategy.

    Parameters
    ----------
    io_config : IOConfigDivCqr
        IO configuration

    Returns
    -------
    BinMatrix
        Binary matrix
    DivCqrStrategyLogger
        Logger

    Raises
    ------
    DivideAndConquerStrategyError
        If the final matrix is degenerated
        or if the solver time limit is reached

    """
    init_bin_matrix = BinMatrix.from_csv(io_config.csv_path())
    logger = DivCqrStrategyLogger(
        io_config,
        BinMatrixStats.from_bin_matrix(init_bin_matrix),
    )

    _LOGGER.info("Input / Output config:\n%s\n", io_config)
    _LOGGER.info(
        "Initial binary matrix stats:\n%s\n",
        str(logger.init_bin_matrix()),
    )

    _LOGGER.info(
        "D&C: solving heuristic problem %s",
        io_config.heuristic_model().value,
    )
    try:
        smaller_bin_matrix, heuristic_solve_stats = solve_heuristic_dsbm(
            init_bin_matrix,
            io_config.heuristic_model(),
            io_config.heuristic_epsilon(),
            io_config.solver(),
        )
    except ModelFindsDegeneratedMatrixError as error:
        _LOGGER.info("Submatrix degenerated, keep original matrix")
        smaller_bin_matrix = init_bin_matrix
        heuristic_solve_stats = error.solve_stats()
    except SolveError as error:
        logger.add_heuristic_solve_stats(error.error().solve_stats())
        _LOGGER.info("Solve stats:\n%s\n", logger.heuristic_solve_stats())
        _LOGGER.critical(str(error))
        raise DivideAndConquerStrategyError(logger) from error

    logger.add_heuristic_solve_stats(heuristic_solve_stats)
    logger.add_heuristic_result_bin_matrix(
        BinMatrixStats.from_bin_matrix(smaller_bin_matrix),
    )
    _LOGGER.info(
        "Heuristic solve stats:\n%s\n",
        logger.heuristic_solve_stats(),
    )
    _LOGGER.info(
        "Heuristic result binary matrix stats:\n%s\n",
        str(logger.heuristic_result_bin_matrix()),
    )
    _LOGGER.info(
        "D&C: solving exact problem %s",
        io_config.exact_model().value,
    )
    try:
        exact_result_bin_matrix, exact_solve_stats = solve_exact_dsbm(
            smaller_bin_matrix,
            io_config.exact_model(),
            io_config.exact_epsilon(),
            io_config.solver(),
        )
    except (ModelFindsDegeneratedMatrixError, SolveError) as error:
        if isinstance(error, ModelFindsDegeneratedMatrixError):
            logger.add_exact_solve_stats(cast(ModelLPMILPStats, error.solve_stats()))
        elif isinstance(error, SolveError):
            logger.add_exact_solve_stats(
                cast(ModelLPMILPStats, error.error().solve_stats()),
            )
        _LOGGER.info("Solve stats:\n%s\n", logger.exact_solve_stats())
        _LOGGER.critical(str(error))
        raise DivideAndConquerStrategyError(logger) from error

    logger.add_exact_solve_stats(exact_solve_stats)
    logger.add_exact_result_bin_matrix(
        BinMatrixStats.from_bin_matrix(exact_result_bin_matrix),
    )
    _LOGGER.info(
        "Exact solve stats:\n%s\n",
        logger.exact_solve_stats(),
    )
    _LOGGER.info(
        "Exact result binary matrix stats:\n%s\n",
        str(logger.exact_result_bin_matrix()),
    )
    return exact_result_bin_matrix, logger


# ============================================================================ #
#                                  EXCEPTIONS                                  #
# ============================================================================ #
class UniqueStrategyError(Exception):
    """Unique strategy exception."""

    def __init__(self, logger: UniqueStrategyLogger) -> None:
        """Initialize the exception."""
        self._LOGGER: UniqueStrategyLogger = logger

    def logger(self) -> UniqueStrategyLogger:
        """Logger.

        Returns
        -------
        UniqueStrategyLogger
            Logger

        """
        return self._LOGGER

    def __str__(self) -> str:
        """Print the exception message."""
        return "Unique strategy error"


class DivideAndConquerStrategyError(Exception):
    """Divide and conquer strategy exception."""

    def __init__(self, logger: DivCqrStrategyLogger) -> None:
        """Initialize the exception."""
        self._LOGGER: DivCqrStrategyLogger = logger

    def logger(self) -> DivCqrStrategyLogger:
        """Logger.

        Returns
        -------
        DivCqrStrategyLogger
            Logger

        """
        return self._LOGGER

    def __str__(self) -> str:
        """Print the exception message."""
        return "Divide and conquer strategy error"



"""Application module for binary matrix views."""

from __future__ import annotations

import logging
from dataclasses import dataclass
from pathlib import Path
from typing import Optional

import yaml  # type: ignore[import-untyped]

try:
    from yaml import CDumper as Dumper
except ImportError:
    from yaml import Dumper

import typer

from qbc_dsbm.binary_matrix import BinMatrix, BinMatrixStats

_LOGGER = logging.getLogger(__name__)
_LOGGER.setLevel(logging.INFO)
console_handler = logging.StreamHandler()
formatter = logging.Formatter(
    "%(asctime)s - %(name)s [%(levelname)s] %(message)s",
)
console_handler.setFormatter(formatter)
_LOGGER.addHandler(console_handler)


@dataclass
class BMStatsArgs:
    """Binary matrix stats args and opts."""

    ARG_INPUT_BM_CSV_PATH = typer.Argument(
        help="Path to the binary matrix CSV file",
    )
    ARG_OUTPUT_DIRECTORY = typer.Option(
        default=None,
        help="Path to the output directory",
    )


APP = typer.Typer()


@APP.command()
def stats(
    binary_matrix_csv: Path = BMStatsArgs.ARG_INPUT_BM_CSV_PATH,
    output_directory: Optional[Path] = BMStatsArgs.ARG_OUTPUT_DIRECTORY,  # noqa: UP007
) -> None:
    """Get stats of a binary matrix."""
    binary_matrix = BinMatrix.from_csv(binary_matrix_csv)
    bin_matrix_stats = BinMatrixStats.from_bin_matrix(binary_matrix)
    _LOGGER.info("Binary matrix:\n%s\n", bin_matrix_stats)

    if output_directory is not None:
        output_directory = Path(output_directory)
        _LOGGER.info("Output directory: %s", output_directory)
        with (output_directory / "bin_matrix_stats.yaml").open("w") as yaml_out:
            yaml_out.write(
                yaml.dump(
                    bin_matrix_stats.to_dict(),
                    Dumper=Dumper,
                    sort_keys=False,
                ),
            )



"""Common logging module."""

from __future__ import annotations

import logging
import sys

__ROOT_APP_NAME = "qbc_dsbm"

_LOGGER = logging.getLogger(__ROOT_APP_NAME)
_LOGGER.setLevel(logging.DEBUG)


def format_logger(
    debug: bool,  # noqa: FBT001
) -> None:
    """Format logger."""
    _LOGGER.handlers.clear()
    _LOGGER.filters.clear()
    handler = logging.StreamHandler(sys.stdout)
    if debug:
        handler.setLevel(logging.DEBUG)
    else:
        handler.setLevel(logging.INFO)
    formatter = logging.Formatter(
        "%(asctime)s %(name)s [%(levelname)s] %(message)s",
    )
    handler.setFormatter(formatter)
    _LOGGER.addHandler(handler)
    log_filter = logging.Filter(__ROOT_APP_NAME)
    _LOGGER.addFilter(log_filter)


def init_logger(
    first_info_message: str,
    debug: bool,  # noqa: FBT001
) -> None:
    """Initialize logger."""
    format_logger(debug)
    _LOGGER.info(first_info_message)



"""Binary matrix operations."""

from qbc_dsbm.binary_matrix import BinMatrix


# ---------------------------------------------------------------------------- #
#                                Row Operations                                #
# ---------------------------------------------------------------------------- #
def a_minus_b_rows(a: BinMatrix, b: BinMatrix) -> BinMatrix:
    """Subtract two binary matrices on the rows.

    Parameters
    ----------
    a : BinMatrix
        A binary matrix
    b : BinMatrix
        A binary matrix

    Returns
    -------
    BinMatrix
        A binary matrix c with the rows of a without the rows of b

    Raises
    ------
    DegeneratedMatrixError
        Degenerated matrix error

    """
    row_mask = [True for _ in range(a.number_of_rows())]
    column_mask = [True for _ in range(a.number_of_columns())]

    a_row_header = a.row_header()
    b_row_header_set = set(b.row_header())

    for k, col_id in enumerate(a_row_header):
        if col_id in b_row_header_set:
            row_mask[k] = False

    return a.to_sub_matrix(row_mask, column_mask)


def a_inter_b_rows(a: BinMatrix, b: BinMatrix) -> BinMatrix:
    """Intersect two binary matrices on the rows.

    Parameters
    ----------
    a : BinMatrix
        A binary matrix
    b : BinMatrix
        A binary matrix

    Returns
    -------
    BinMatrix
        A binary matrix c with the rows in common in a and b

    Raises
    ------
    DegeneratedMatrixError
        Degenerated matrix error

    """
    row_mask = [False for _ in range(a.number_of_rows())]
    column_mask = [True for _ in range(a.number_of_columns())]

    a_row_header = a.row_header()
    b_row_header_set = set(b.row_header())

    for k, row_id in enumerate(a_row_header):
        if row_id in b_row_header_set:
            row_mask[k] = True

    return a.to_sub_matrix(row_mask, column_mask)


# ---------------------------------------------------------------------------- #
#                               Column Operations                              #
# ---------------------------------------------------------------------------- #
def a_minus_b_columns(a: BinMatrix, b: BinMatrix) -> BinMatrix:
    """Subtract two binary matrices on the columns.

    Parameters
    ----------
    a : BinMatrix
        A binary matrix
    b : BinMatrix
        A binary matrix

    Returns
    -------
    BinMatrix
        A binary matrix c with the columns of a without the columns of b

    Raises
    ------
    DegeneratedMatrixError
        Degenerated matrix error

    """
    row_mask = [True for _ in range(a.number_of_rows())]
    column_mask = [True for _ in range(a.number_of_columns())]

    a_col_header = a.column_header()
    b_col_header_set = set(b.column_header())

    for k, col_id in enumerate(a_col_header):
        if col_id in b_col_header_set:
            column_mask[k] = False

    return a.to_sub_matrix(row_mask, column_mask)


def a_inter_b_columns(a: BinMatrix, b: BinMatrix) -> BinMatrix:
    """Intersect two binary matrices on the columns.

    Parameters
    ----------
    a : BinMatrix
        A binary matrix
    b : BinMatrix
        A binary matrix

    Returns
    -------
    BinMatrix
        A binary matrix c with the columns in common in a and b

    Raises
    ------
    DegeneratedMatrixError
        Degenerated matrix error

    """
    row_mask = [True for _ in range(a.number_of_rows())]
    column_mask = [False for _ in range(a.number_of_columns())]

    a_col_header = a.column_header()
    b_col_header_set = set(b.column_header())

    for k, col_id in enumerate(a_col_header):
        if col_id in b_col_header_set:
            column_mask[k] = True

    return a.to_sub_matrix(row_mask, column_mask)



"""Common argument module."""

# Due to typer usage:
# ruff: noqa: TCH001, TCH003, UP007, FBT001, PLR0913

from __future__ import annotations

from dataclasses import dataclass

import typer


@dataclass
class CommonArgs:
    """Common args and opts."""

    OPT_DEBUG = typer.Option(
        help="Debug mode",
    )



"""Init file for module path shortcuts."""

from __future__ import annotations

from enum import Enum


# ============================================================================ #
#                                     TYPES                                    #
# ============================================================================ #
class QBCExactModels(Enum):
    """QBC exact model enum."""

    MAX_ONES = "max_ones"
    MAX_SURFACE = "max_surface"
    MAX_PERIMETER = "max_perimeter"
    MAX_ONES_COMPACT = "max_ones_compact"


class BCModels(Enum):
    """BC model enum."""

    KONIG_V = "König_V"
    KONIG_E = "König_E"


class QBCHeuristicModels(Enum):
    """QBC heuristic model enum."""

    MIN_DEL_RC = "min_del_rc"
    MIN_DEL_ROWS = "min_del_rows"
    MIN_DEL_COLS = "min_del_cols"
    MIN_DEL_ROWS_RELAX = "min_del_rows_relax"
    MIN_DEL_COLS_RELAX = "min_del_cols_relax"
    MIN_DEL_ONES = "min_del_ones"
    KP_QB = "KP_QB"


UniqueModels = QBCExactModels | BCModels | QBCHeuristicModels


# ============================================================================ #
#                                  EXCEPTIONS                                  #
# ============================================================================ #
class InvalidModelNameError(Exception):
    """Invalid model name."""

    def __init__(self, invalid_model_str: str) -> None:
        """Initialize exception."""
        self.__invalid_model_str = invalid_model_str

    def invalid_model_str(self) -> str:
        """Invalid model string.

        Returns
        -------
        str
            Invalid model string

        """
        return self.__invalid_model_str

    def __str__(self) -> str:
        """Give the string representation."""
        return f"Invalid model name: {self.__invalid_model_str}"


def string_to_model_enum(model_str: str) -> UniqueModels:
    """Convert string to model enum.

    Parameters
    ----------
    model_str : str
        Model string

    Returns
    -------
    UniqueModels
        Model enum

    Raises
    ------
    InvalidModelNameError
        Invalid model name

    """
    if model_str in QBCExactModels.__members__:
        return QBCExactModels[model_str]
    if model_str in BCModels.__members__:
        return BCModels[model_str]
    if model_str in QBCHeuristicModels.__members__:
        return QBCHeuristicModels[model_str]
    raise InvalidModelNameError(model_str)
